# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import pytest
import torch

from cellarium.ml.preprocessing.highly_variable_genes import get_highly_variable_genes


@pytest.mark.parametrize("num_genes_to_check", [2, 3, 4])
def test_highly_variable_genes_top_n(num_genes_to_check: int):
    """
    Test if :func:`get_highly_variable_genes` returns the top n genes.
    """
    # data
    gene_names = ["gene_0", "gene_1", "gene_2", "gene_3", "gene_4", "gene_5"]
    mean = torch.tensor([1, 2, 3, 4, 5, 6])
    var = torch.tensor(
        [
            1.6086e00,
            2.2582e02,
            9.8922e-02,
            1.4379e01,
            4.0901e-02,
            9.9200e-01,
        ]
    )
    # compute
    result = get_highly_variable_genes(
        gene_names=gene_names, mean=mean, var=var, n_top_genes=num_genes_to_check, n_bins=1
    )
    # assert
    assert result[result.highly_variable].shape[0] == num_genes_to_check


def test_highly_variable_genes_cutoffs():
    """
    Test if :func:`get_highly_variable_genes` returns the genes within the cutoffs.
    """
    # data
    gene_names = ["gene_0", "gene_1", "gene_2", "gene_3", "gene_4", "gene_5"]
    mean = torch.tensor([1, 2, 3, 4, 5, 6])
    var = torch.tensor(
        [
            1.6086e00,
            2.2582e02,
            9.8922e-02,
            1.4379e01,
            4.0901e-02,
            9.9200e-01,
        ]
    )
    # compute
    result = get_highly_variable_genes(
        gene_names, mean, var, min_disp=0.2, max_disp=10, min_mean=0, max_mean=5, n_bins=1
    )
    # assert
    assert result[result.highly_variable].shape[0] != 0


def test_highly_variable_genes_wrong_sizes():
    """
    Test if :func:`get_highly_variable_genes` raises an error when the sizes of the inputs are different.
    """
    # data
    gene_names = ["gene_0", "gene_1", "gene_2", "gene_3"]
    mean = torch.tensor([1, 2, 3])
    var = torch.tensor([0.1, 0.2, 0.3, 0.4])
    # assert
    with pytest.raises(ValueError):
        get_highly_variable_genes(gene_names, mean, var)



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Sequence

import numpy as np
import torch
from torch import nn

from cellarium.ml.utilities.testing import (
    assert_arrays_equal,
    assert_columns_and_array_lengths_equal,
    assert_nonnegative,
)


class ZScore(nn.Module):
    """
    ZScore gene counts with  mean and standard deviation.

    .. math::

        y_{ng} = \\frac{x_{ng} - \\mathrm{mean}_g}{\\mathrm{std}_g + \\mathrm{eps}}

    Args:
        mean_g:
            Means for each gene.
        std_g:
            Standard deviations for each gene.
        var_names_g:
            The variable names schema for the input data validation.
        eps:
            A value added to the denominator for numerical stability.
    """

    def __init__(
        self,
        mean_g: torch.Tensor,
        std_g: torch.Tensor,
        var_names_g: Sequence[str],
        eps: float = 1e-6,
    ) -> None:
        super().__init__()
        self.mean_g: torch.Tensor
        self.std_g: torch.Tensor
        self.register_buffer("mean_g", mean_g)
        self.register_buffer("std_g", std_g)
        self.var_names_g = np.array(var_names_g)
        assert_nonnegative("eps", eps)
        self.eps = eps

    def forward(
        self,
        x_ng: torch.Tensor,
        var_names_g: np.ndarray,
    ) -> dict[str, torch.Tensor]:
        """
        .. note::

            When used with :class:`~cellarium.ml.core.CellariumModule` or :class:`~cellarium.ml.core.CellariumPipeline`,
            ``x_ng`` key in the input dictionary will be overwritten with the z-scored values.

        Args:
            x_ng:
                Gene counts.
            var_names_g:
                The list of the variable names in the input data. If ``None``, no validation is performed.

        Returns:
            A dictionary with the following keys:

            - ``x_ng``: The z-scored gene counts.
        """
        assert_columns_and_array_lengths_equal("x_ng", x_ng, "var_names_g", var_names_g)
        assert_arrays_equal("var_names_g", var_names_g, "var_names_g", self.var_names_g)

        x_ng = (x_ng - self.mean_g) / (self.std_g + self.eps)
        return {"x_ng": x_ng}

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(mean_g={self.mean_g}, std_g={self.std_g}, "
            f"var_names_g={self.var_names_g}), eps={self.eps}"
        )



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

"""
Testing utilities
-----------------

This module contains helper functions for testing.
"""

from collections.abc import Callable
from typing import Any

import numpy as np
import pandas as pd
import torch
from pytest import approx
from scipy.stats import linregress


def assert_positive(name: str, number: float):
    """
    Assert that a number is positive.

    Args:
        name: The name of the number.
        number: The number to check.

    Raises:
        ValueError: If the number is not positive.
    """
    if number <= 0:
        raise ValueError(f"`{name}` must be positive. Got {number}")


def assert_nonnegative(name: str, number: float):
    """
    Assert that a number is non-negative.

    Args:
        name: The name of the number.
        number: The number to check.

    Raises:
        ValueError: If the number is negative.
    """
    if number < 0:
        raise ValueError(f"`{name}` must be non-negative. Got {number}")


def assert_columns_and_array_lengths_equal(
    matrix_name: str,
    matrix: np.ndarray | torch.Tensor,
    array_name: str,
    array: np.ndarray | torch.Tensor,
):
    """
    Assert that the number of columns in a matrix matches the length of an array.

    Args:
        matrix_name: The name of the matrix.
        matrix: The matrix.
        array_name: The name of the array.
        array: The array.

    Raises:
        ValueError: If the number of columns in the matrix does not match the length of the array.
    """
    if matrix.shape[1] != len(array):
        raise ValueError(
            f"The number of `{matrix_name}` columns must match the `{array_name}` length. "
            f"Got {matrix.shape[1]} != {len(array)}"
        )


def assert_arrays_equal(
    a1_name: str,
    a1: np.ndarray,
    a2_name: str,
    a2: np.ndarray,
):
    """
    Assert that two arrays are equal.

    Args:
        a1_name: The name of the first array.
        a1: The first array.
        a2_name: The name of the second array.
        a2: The second array.

    Raises:
        ValueError: If the arrays are not equal.
    """
    if not np.array_equal(a1, a2):
        raise ValueError(f"`{a1_name}` must match `{a2_name}`. " f"Got {a1} != {a2}")


def assert_slope_equals(data: pd.Series, slope: float, loglog: bool = False, atol: float = 1e-4):
    """
    Assert that the slope of a series is equal to a given value.

    Args:
        data:
            The :class:`pandas.Series` object to check.
        slope:
            Expected slope.
        loglog:
            Whether to use log-log scale.
        atol:
            The absolute tolerance.

    Raises:
        ValueError: If the slope is not equal to the given value.
    """
    x = np.log(data.index) if loglog else data.index
    y = np.log(data) if loglog else data
    assert linregress(x, y).slope == approx(slope, abs=atol)


"""
Helper functions for coordinate data collection.
"""


def l1_norm(x: torch.Tensor) -> float:
    return x.detach().abs().mean().item()


def record_out_coords(
    records: list[dict], width: int, name: str, t: int
) -> Callable[[torch.nn.Module, torch.Tensor, torch.Tensor], None]:
    """
    Returns a hook to record layer output coordinate size.

    Args:
        records:
            The list of records to append to.
        width:
            The width of the model.
        name:
            The name of the layer.
        t:
            The time step.

    Returns:
        A hook to record layer output coordinate size.
    """

    def hook(module: torch.nn.Module, input: torch.Tensor, output: torch.Tensor) -> None:
        ret = {"width": width, "module": f"{name}.out", "t": t, "l1": l1_norm(output), "type": "out"}
        records.append(ret)

    return hook


def get_coord_data(
    models: dict[int, Callable[[], torch.nn.Module]],
    train_loader: torch.utils.data.DataLoader,
    loss_fn: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    optim_fn: type[torch.optim.Optimizer],
    lr: float,
    nsteps: int,
    nseeds: int,
) -> pd.DataFrame:
    """
    Get coordinate data for a model.

    Args:
        models:
            A dictionary mapping width to a function that returns a model.
        train_loader:
            The training data loader.
        loss_fn:
            The loss function.
        optim_fn:
            The optimizer class.
        lr:
            The learning rate.
        nsteps:
            The number of steps to train for.
        nseeds:
            The number of seeds to use.

    Returns:
        A :class:`pandas.DataFrame` containing the coordinate data.
    """
    records: list[dict[str, object]] = []
    for i in range(nseeds):
        torch.manual_seed(i)
        for width, lazy_model in models.items():
            model = lazy_model()
            model.train()
            optim_kwargs: dict[str, Any] = {"lr": lr}
            optimizer = optim_fn(model.parameters(), **optim_kwargs)
            data_iter = iter(train_loader)

            for batch_idx in range(nsteps):
                data, target = next(data_iter)
                remove_hooks = []
                prev_param = {}
                for module_name, module in model.named_children():
                    # record layer outputs
                    remove_hooks.append(
                        module.register_forward_hook(record_out_coords(records, width, module_name, batch_idx))  # type: ignore[arg-type]
                    )
                    # record parameter values
                    for param_name, param in module.named_parameters():
                        if param_name.endswith("_unscaled"):
                            # muP
                            param_name = param_name.removesuffix("_unscaled")
                            multiplier = getattr(module, f"{param_name}_multiplier")
                        else:
                            # SP
                            multiplier = 1.0
                        ret = {
                            "width": width,
                            "module": f"{module_name}.{param_name}",
                            "t": batch_idx,
                            "l1": multiplier * l1_norm(param),
                            "type": "param",
                        }
                        records.append(ret)
                        prev_param[f"{module_name}.{param_name}"] = param.detach().clone()

                # step
                optimizer.zero_grad()
                output = model(data.view(data.size(0), -1))
                loss = loss_fn(output, target)
                loss.backward()
                optimizer.step()

                # record parameter deltas
                for module_name, module in model.named_children():
                    for param_name, param in module.named_parameters():
                        if param_name.endswith("_unscaled"):
                            # muP
                            param_name = param_name.removesuffix("_unscaled")
                            multiplier = getattr(module, f"{param_name}_multiplier")
                        else:
                            # SP
                            multiplier = 1.0
                        delta_param = param.detach() - prev_param[f"{module_name}.{param_name}"]
                        ret = {
                            "width": width,
                            "module": f"{module_name}.{param_name}.delta",
                            "t": batch_idx,
                            "l1": multiplier * l1_norm(delta_param),
                            "type": "delta",
                        }
                        records.append(ret)

                for handle in remove_hooks:
                    handle.remove()

    return pd.DataFrame(records)



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np
import torch

from cellarium.ml.models import PredictMixin, ValidateMixin
from cellarium.ml.utilities.core import call_func_with_batch


class CellariumPipeline(torch.nn.ModuleList):
    """
    A pipeline of modules. Modules are expected to return a dictionary. The input dictionary is sequentially passed to
    (piped through) each module and updated with its output dictionary.

    When used within :class:`cellarium.ml.core.CellariumModule`, the last module in the pipeline is expected to be
    a model (:class:`cellarium.ml.models.CellariumModel`) and any preceding modules are expected to be data transforms.

    Example:

        >>> from cellarium.ml import CellariumPipeline
        >>> from cellarium.ml.transforms import NormalizeTotal, Log1p
        >>> from cellarium.ml.models import IncrementalPCA
        >>> pipeline = CellariumPipeline([
        ...     NormalizeTotal(),
        ...     Log1p(),
        ...     IncrementalPCA(var_names_g=[f"gene_{i}" for i in range(20)], n_components=10),
        ... ])
        >>> batch = {"x_ng": x_ng, "total_mrna_umis_n": total_mrna_umis_n, "var_names_g": var_names_g}
        >>> output = pipeline(batch)  # or pipeline.predict(batch)

    Args:
        modules:
            Modules to be executed sequentially.
    """

    def forward(self, batch: dict[str, np.ndarray | torch.Tensor]) -> dict[str, torch.Tensor | np.ndarray]:
        for module in self:
            batch |= call_func_with_batch(module.forward, batch)

        return batch

    def predict(self, batch: dict[str, np.ndarray | torch.Tensor]) -> dict[str, np.ndarray | torch.Tensor]:
        for module in self[:-1]:
            batch |= call_func_with_batch(module.forward, batch)

        model = self[-1]
        if not isinstance(model, PredictMixin):
            raise TypeError(f"The last module in the pipeline must be an instance of {PredictMixin}. Got {model}")
        batch |= call_func_with_batch(model.predict, batch)

        return batch

    def validate(self, batch: dict[str, np.ndarray | torch.Tensor]) -> None:
        for module in self[:-1]:
            batch |= call_func_with_batch(module.forward, batch)

        model = self[-1]
        if not isinstance(model, ValidateMixin):
            raise TypeError(f"The last module in the pipeline must be an instance of {ValidateMixin}. Got {model}")
        call_func_with_batch(model.validate, batch)



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import math
from collections.abc import Callable
from pathlib import Path
from typing import Literal

import pandas as pd
import pytest
import torch
import torch.nn.functional as F
from torch import nn
from torchvision import datasets, transforms

from cellarium.ml.models import MuLinear
from cellarium.ml.utilities.testing import assert_slope_equals, get_coord_data

optim_dict = {"sgd": torch.optim.SGD, "adam": torch.optim.Adam, "adamw": torch.optim.AdamW}


# adapted from https://github.com/microsoft/mup/blob/main/examples/MLP/main.py
def coord_check_MLP(
    mup: bool,
    bias: bool,
    nonlin: Callable[[torch.Tensor], torch.Tensor],
    lr: float,
    input_mult: float,
    output_mult: float,
    optim_name: Literal["sgd", "adam", "adamw"],
    train_loader: torch.utils.data.DataLoader,
    nsteps: int,
    nseeds: int,
    widths: list[int],
) -> pd.DataFrame:
    """
    Coordinate checking for an MLP model.

    Args:
        mup:
            If ``True``, use the μP MLP model. Otherwise, use the SP MLP model.
        bias:
            If ``True``, use bias in the model.
        nonlin:
            Nonlinearity function to use in the model.
        lr:
            Learning rate.
        input_mult:
            Input multiplier.
        output_mult:
            Output multiplier.
        optim_name:
            Name of the optimizer to use.
        train_loader:
            Training data loader.
        nsteps:
            Number of steps to train for.
        nseeds:
            Number of repeats with different random seeds.
        widths:
            List of widths of the models to use for coordinate checking.

    Returns:
        Dataframe with the results of the coordinate checking.
    """

    def gen(w: int) -> Callable[[], nn.Module]:
        def f() -> nn.Module:
            model: nn.Module
            if mup:
                model = MuMLP(
                    width=w,
                    bias=bias,
                    nonlin=nonlin,
                    optimizer=optim_name,
                    input_mult=input_mult,
                    output_mult=output_mult,
                )
            else:
                model = MLP(width=w, bias=bias, nonlin=nonlin, input_mult=input_mult, output_mult=output_mult)
            return model

        return f

    models = {w: gen(w) for w in widths}
    optim_fn = optim_dict[optim_name]

    return get_coord_data(
        models,
        train_loader,
        loss_fn=F.cross_entropy,
        lr=lr,
        optim_fn=optim_fn,
        nseeds=nseeds,
        nsteps=nsteps,
    )


@pytest.fixture
def train_loader(tmp_path: Path) -> torch.utils.data.DataLoader:
    batch_size = 64
    transform = transforms.Compose([transforms.ToTensor(), transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))])

    trainset = datasets.CIFAR10(root=tmp_path, train=True, download=True, transform=transform)
    return torch.utils.data.DataLoader(trainset, batch_size=batch_size, shuffle=True)


class MLP(nn.Module):
    """
    A Multi-Layer Perceptron (MLP) model with 3 layers in standard parameterization (SP).

    Args:
        width:
            Width of the hidden layer.
        num_classes:
            Number of classes.
        bias:
            If ``True``, use bias in the model.
        nonlin:
            Nonlinearity function to use in the model.
        input_mult:
            Input multiplier.
        output_mult:
            Output multiplier.
    """

    def __init__(
        self,
        width: int = 128,
        num_classes: int = 10,
        bias: bool = False,
        nonlin: Callable[[torch.Tensor], torch.Tensor] = F.relu,
        input_mult: float = 1.0,
        output_mult: float = 1.0,
    ) -> None:
        super().__init__()
        self.nonlin = nonlin
        self.bias = bias
        self.input_mult = input_mult
        self.output_mult = output_mult
        self.fc_1 = nn.Linear(3072, width, bias=bias)
        self.fc_2 = nn.Linear(width, width, bias=bias)
        self.fc_3 = nn.Linear(width, num_classes, bias=bias)
        self.reset_parameters()

    def reset_parameters(self) -> None:
        fan_in_1 = self.fc_1.weight.shape[1]
        nn.init.normal_(self.fc_1.weight, std=1 / math.sqrt(fan_in_1))  # 1 / sqrt(d)
        self.fc_1.weight.data /= self.input_mult**0.5
        fan_in_2 = self.fc_2.weight.shape[1]
        nn.init.normal_(self.fc_2.weight, std=1 / math.sqrt(fan_in_2))  # 1 / sqrt(n)
        nn.init.zeros_(self.fc_3.weight)  # zero readout
        if self.bias:
            # zero biases
            nn.init.zeros_(self.fc_1.bias)
            nn.init.zeros_(self.fc_2.bias)
            nn.init.zeros_(self.fc_3.bias)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.nonlin(self.fc_1(x) * self.input_mult**0.5)
        x = self.nonlin(self.fc_2(x))
        return self.fc_3(x) * self.output_mult


class MuMLP(nn.Module):
    """
    A Multi-Layer Perceptron (MLP) model with 3 layers in maximal update parameterization (μP).

    Args:
        width:
            Width of the hidden layer.
        num_classes:
            Number of classes.
        bias:
            If ``True``, use bias in the model.
        nonlin:
            Nonlinearity function to use in the model.
        optimizer:
            Name of the optimizer to use.
        input_mult:
            Input multiplier.
        output_mult:
            Output multiplier.
    """

    def __init__(
        self,
        width: int = 128,
        num_classes: int = 10,
        bias: bool = False,
        nonlin: Callable[[torch.Tensor], torch.Tensor] = F.relu,
        optimizer: Literal["sgd", "adam", "adamw"] = "sgd",
        input_mult: float = 1.0,
        output_mult: float = 1.0,
    ) -> None:
        super().__init__()
        self.nonlin = nonlin
        self.input_mult = input_mult
        self.output_mult = output_mult
        self.fc_1 = MuLinear(
            in_features=3072,
            out_features=width,
            bias=bias,
            layer="input",
            optimizer=optimizer,
            weight_init_std=(1 / math.sqrt(3072 * self.input_mult)),
            base_width=128,
        )
        self.fc_2 = MuLinear(
            in_features=width,
            out_features=width,
            bias=bias,
            layer="hidden",
            optimizer=optimizer,
            weight_init_std=(1 / math.sqrt(128)),
            base_width=128,
        )
        self.fc_3 = MuLinear(
            in_features=width,
            out_features=num_classes,
            bias=bias,
            layer="output",
            optimizer=optimizer,
            weight_init_std=0.0,
            base_width=128,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.nonlin(self.fc_1(x) * self.input_mult**0.5)
        x = self.nonlin(self.fc_2(x))
        return self.fc_3(x) * self.output_mult


@pytest.mark.parametrize("bias", [True, False])
@pytest.mark.parametrize("nonlin", [F.relu, F.tanh])
@pytest.mark.parametrize(
    "optimizer,lr,input_mult,output_mult",
    [("sgd", 0.1, 2**-8, 2**5), ("adam", 0.01, 2**-6, 2**-4), ("adamw", 0.01, 2**-6, 2**-4)],
)
def test_mup(
    train_loader: torch.utils.data.DataLoader,
    optimizer: Literal["sgd", "adam", "adamw"],
    lr: float,
    input_mult: float,
    output_mult: float,
    nonlin: Callable[[torch.Tensor], torch.Tensor],
    bias: bool,
):
    """
    Perform coordinate checking for the μP MLP model.


    Coordinate scaling of pre-activation layers, model weights, and changes in weights with respect to width
    are described in the Appendix J.2 of [1].

    **References:**

    1. `Tensor Programs V: Tuning Large Neural Networks via Zero-Shot Hyperparameter Transfer (Yang et al.)
       <https://arxiv.org/abs/2203.03466.pdf>`_.
    """
    nsteps = 3
    nseeds = 5
    widths = [2**i for i in range(7, 14)]
    df = coord_check_MLP(
        mup=True,
        bias=bias,
        nonlin=nonlin,
        lr=lr,
        input_mult=input_mult,
        output_mult=output_mult,
        optim_name=optimizer,
        train_loader=train_loader,
        nsteps=nsteps,
        nseeds=nseeds,
        widths=widths,
    )
    atol = 0.1
    for t in range(nsteps):
        df_t = df.loc[df.t == t]
        for i in range(1, 4):
            df_i = df_t.loc[df_t.module.str.contains(str(i))]
            # output scaling: Θ(1)
            df_out = df_i.loc[df.type == "out"]
            out = df_out.groupby("width").l1.mean()
            if any(out):
                assert_slope_equals(out, 0, loglog=True, atol=atol)

            # W
            df_param = df_i.loc[df.type == "param"]
            W = df_param[df_param.module.str.contains("weight")].groupby("width").l1.mean()
            if i == 1:
                # input layer scaling: Θ(1)
                assert_slope_equals(W, 0, loglog=True, atol=atol)
            elif i == 2:
                # hidden layer scaling: Θ(1/sqrt(n))
                assert_slope_equals(W, -0.5, loglog=True, atol=atol)
            elif i == 3:
                # output layer scaling: Θ(1/n)
                if any(W):
                    assert_slope_equals(W, -1, loglog=True, atol=atol)

            # b scaling: Θ(1)
            if bias:
                b = df_param[df_param.module.str.contains("bias")].groupby("width").l1.mean()
                if any(b):
                    assert_slope_equals(b, 0, loglog=True, atol=atol)

            # ∆W
            df_delta = df_i.loc[df.type == "delta"]
            dW = df_delta[df_delta.module.str.contains("weight")].groupby("width").l1.mean()
            if i == 1:
                # input layer scaling: Θ(1)
                if any(dW):
                    assert_slope_equals(dW, 0, loglog=True, atol=atol)
            elif i == 2:
                # hidden layer scaling: Θ(1/n)
                if any(dW):
                    if optimizer == "adamw":
                        # additional scaling due weight decay: Θ(1/sqrt(n))
                        assert_slope_equals(dW, -0.75, loglog=True, atol=0.27)
                    else:
                        assert_slope_equals(dW, -1, loglog=True, atol=atol)
            elif i == 3:
                # output layer scaling: Θ(1/n)
                assert_slope_equals(dW, -1, loglog=True, atol=atol)

            # ∆b scaling: Θ(1)
            if bias:
                db = df_delta[df_delta.module.str.contains("bias")].groupby("width").l1.mean()
                if any(db):
                    assert_slope_equals(db, 0, loglog=True, atol=atol)

    with pytest.raises(AssertionError):
        df = coord_check_MLP(
            mup=False,
            bias=bias,
            nonlin=nonlin,
            lr=lr,
            input_mult=input_mult,
            output_mult=output_mult,
            optim_name=optimizer,
            train_loader=train_loader,
            nsteps=nsteps,
            nseeds=nseeds,
            widths=widths,
        )
        atol = 0.2
        for t in range(nsteps):
            df_t = df.loc[df.t == t]
            for i in range(1, 4):
                df_i = df_t.loc[df_t.module.str.contains(str(i))]
                # output scaling: Θ(1)
                df_out = df_i.loc[df.type == "out"]
                out = df_out.groupby("width").l1.mean()
                if any(out):
                    assert_slope_equals(out, 0, loglog=True, atol=atol)

                # W
                df_param = df_i.loc[df.type == "param"]
                W = df_param[df_param.module.str.contains("weight")].groupby("width").l1.mean()
                if i == 1:
                    # input layer scaling: Θ(1)
                    assert_slope_equals(W, 0, loglog=True, atol=atol)
                elif i == 2:
                    # hidden layer scaling: Θ(1/sqrt(n))
                    assert_slope_equals(W, -0.5, loglog=True, atol=atol)
                elif i == 3:
                    # output layer scaling: Θ(1/n)
                    if any(W):
                        assert_slope_equals(W, -1, loglog=True, atol=atol)

                # b scaling: Θ(1)
                if bias:
                    b = df_param[df_param.module.str.contains("bias")].groupby("width").l1.mean()
                    if any(b):
                        assert_slope_equals(b, 0, loglog=True, atol=0.05)

                # ∆W
                df_delta = df_i.loc[df.type == "delta"]
                dW = df_delta[df_delta.module.str.contains("weight")].groupby("width").l1.mean()
                if i == 1:
                    # input layer scaling: Θ(1)
                    if any(dW):
                        assert_slope_equals(dW, 0, loglog=True, atol=atol)
                elif i == 2:
                    # hidden layer scaling: Θ(1/n)
                    if any(dW):
                        if optimizer == "adamw":
                            # additional scaling due weight decay: Θ(1/sqrt(n))
                            assert_slope_equals(dW, -0.75, loglog=True, atol=0.27)
                        else:
                            assert_slope_equals(dW, -1, loglog=True, atol=atol)
                elif i == 3:
                    # output layer scaling: Θ(1/n)
                    assert_slope_equals(dW, -1, loglog=True, atol=atol)

                # ∆b scaling: Θ(1)
                if bias:
                    db = df_delta[df_delta.module.str.contains("bias")].groupby("width").l1.mean()
                    if any(db):
                        assert_slope_equals(db, 0, loglog=True, atol=atol)



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from cellarium.ml.core import CellariumAnnDataDataModule, CellariumModule, CellariumPipeline

__all__ = ["CellariumAnnDataDataModule", "CellariumModule", "CellariumPipeline"]



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from numbers import Number

import torch
from pyro.distributions import TorchDistribution, constraints
from torch.distributions.utils import broadcast_all, lazy_property


class NegativeBinomial(TorchDistribution):
    """Negative binomial distribution.

    Args:
        mu:
            Mean of the distribution.
        theta:
            Inverse dispersion.
    """

    THETA_THRESHOLD_STIRLING_SWITCH = 200

    arg_constraints = {"mu": constraints.greater_than_eq(0), "theta": constraints.greater_than_eq(0)}
    support = constraints.nonnegative_integer

    def __init__(self, mu: torch.Tensor, theta: torch.Tensor, validate_args: bool | None = None) -> None:
        self.mu, self.theta = broadcast_all(mu, theta)
        if isinstance(mu, Number) and isinstance(theta, Number):
            batch_shape = torch.Size()
        else:
            batch_shape = self.mu.size()
        super().__init__(batch_shape, validate_args=validate_args)

    @property
    def mean(self) -> torch.Tensor:
        return self.mu

    @property
    def variance(self) -> torch.Tensor:
        return (self.mu + (self.mu**2) / self.theta).masked_fill(self.theta == 0, 0)

    @lazy_property
    def _gamma(self) -> torch.distributions.Gamma:
        # Note we avoid validating because self.theta can be zero.
        return torch.distributions.Gamma(
            concentration=self.theta,
            rate=(self.theta / self.mu).masked_fill(self.theta == 0, 1.0),
            validate_args=False,
        )

    def sample(self, sample_shape=torch.Size()) -> torch.Tensor:
        with torch.no_grad():
            rate = self._gamma.sample(sample_shape=sample_shape)
            return torch.poisson(rate)

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:  # type: ignore[override]
        if self._validate_args:
            self._validate_sample(value)

        # Original implementation from scVI:
        #
        #   log_theta_mu_eps = torch.log(self.theta + self.mu + self.eps)
        #   return (
        #       self.theta * (torch.log(self.theta + self.eps) - log_theta_mu_eps)
        #       + value * (torch.log(self.mu + self.eps) - log_theta_mu_eps)
        #       + torch.lgamma(value + self.theta)
        #       - torch.lgamma(self.theta)
        #       - torch.lgamma(value + 1)
        #   )
        delta = torch.where(
            self.theta > self.THETA_THRESHOLD_STIRLING_SWITCH,
            (value + self.theta - 0.5) * torch.log1p(value / self.theta) - value,
            (value + self.theta).lgamma() - self.theta.lgamma() - torch.xlogy(value, self.theta),
        )
        # The case self.theta == 0 and value == 0 has probability 1.
        # The case self.theta == 0 and value != 0 has probability 0.
        return (
            (delta - (value + self.theta) * torch.log1p(self.mu / self.theta)).masked_fill(self.theta == 0, 0)
            - (value + 1).lgamma()
            + torch.xlogy(value, self.mu)
        )



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Sequence

import numpy as np
import torch
from transformers import BertConfig, BertForMaskedLM

from cellarium.ml.models.model import CellariumModel, PredictMixin
from cellarium.ml.utilities.testing import (
    assert_arrays_equal,
    assert_columns_and_array_lengths_equal,
)


class Geneformer(CellariumModel, PredictMixin):
    """
    Geneformer model.

    **References:**

    1. `Transfer learning enables predictions in network biology (Theodoris et al.)
       <https://www.nature.com/articles/s41586-023-06139-9>`_.

    Args:
        var_names_g:
            The variable names schema for the input data validation.
        hidden_size:
            Dimensionality of the encoder layers and the pooler layer.
        num_hidden_layers:
            Number of hidden layers in the Transformer encoder.
        num_attention_heads:
            Number of attention heads for each attention layer in the Transformer encoder.
        intermediate_size:
            Dimensionality of the "intermediate" (often named feed-forward) layer in the Transformer encoder.
        hidden_act:
            The non-linear activation function (function or string) in the encoder and pooler. If string, ``"gelu"``,
            ``"relu"``, ``"silu"`` and ``"gelu_new"`` are supported.
        hidden_dropout_prob:
            The dropout probability for all fully connected layers in the embeddings, encoder, and pooler.
        attention_probs_dropout_prob:
            The dropout ratio for the attention probabilities.
        max_position_embeddings:
            The maximum sequence length that this model might ever be used with. Typically set this to something large
            just in case (e.g., 512 or 1024 or 2048).
        type_vocab_size:
            The vocabulary size of the ``token_type_ids`` passed when calling :class:`transformers.BertModel`.
        initializer_range:
            The standard deviation of the truncated_normal_initializer for initializing all weight matrices.
        position_embedding_type:
            Type of position embedding. Choose one of ``"absolute"``, ``"relative_key"``, ``"relative_key_query"``. For
            positional embeddings use ``"absolute"``. For more information on ``"relative_key"``, please refer to
            `Self-Attention with Relative Position Representations (Shaw et al.) <https://arxiv.org/abs/1803.02155>`_.
            For more information on ``"relative_key_query"``, please refer to *Method 4* in `Improve Transformer Models
            with Better Relative Position Embeddings (Huang et al.) <https://arxiv.org/abs/2009.13658>`_.
        layer_norm_eps:
            The epsilon used by the layer normalization layers.
        mlm_probability:
            Ratio of tokens to mask for masked language modeling loss.
    """

    def __init__(
        self,
        var_names_g: Sequence[str],
        hidden_size: int = 256,
        num_hidden_layers: int = 6,
        num_attention_heads: int = 4,
        intermediate_size: int = 512,
        hidden_act: str = "relu",
        hidden_dropout_prob: float = 0.02,
        attention_probs_dropout_prob: float = 0.02,
        max_position_embeddings: int = 2048,
        type_vocab_size: int = 2,
        initializer_range: float = 0.02,
        position_embedding_type: str = "absolute",
        layer_norm_eps: float = 1e-12,
        mlm_probability: float = 0.15,
    ) -> None:
        super().__init__()
        self.var_names_g = np.array(var_names_g)
        # model configuration
        config = {
            "vocab_size": len(self.var_names_g) + 2,  # number of genes + 2 for <mask> and <pad> tokens
            "hidden_size": hidden_size,
            "num_hidden_layers": num_hidden_layers,
            "num_attention_heads": num_attention_heads,
            "intermediate_size": intermediate_size,
            "hidden_act": hidden_act,
            "attention_probs_dropout_prob": attention_probs_dropout_prob,
            "hidden_dropout_prob": hidden_dropout_prob,
            "max_position_embeddings": max_position_embeddings,
            "type_vocab_size": type_vocab_size,
            "initializer_range": initializer_range,
            "position_embedding_type": position_embedding_type,
            "layer_norm_eps": layer_norm_eps,
            "pad_token_id": 0,
        }
        self.config = BertConfig(**config)
        self.bert = BertForMaskedLM(self.config)
        self.mlm_probability = mlm_probability
        self.feature_ids: torch.Tensor
        # ids for the features, 0 is for padding, 1 is for mask
        self.register_buffer("feature_ids", torch.arange(2, len(self.var_names_g) + 2))
        self.reset_parameters()

    def reset_parameters(self) -> None:
        self.feature_ids = torch.arange(2, len(self.var_names_g) + 2)
        self.bert.bert.embeddings.position_ids = torch.arange(self.config.max_position_embeddings).expand((1, -1))
        self.bert.bert.embeddings.token_type_ids = torch.zeros(
            self.bert.bert.embeddings.position_ids.size(), dtype=torch.long
        )
        self.bert.apply(lambda module: setattr(module, "_is_hf_initialized", False))
        self.bert.init_weights()

    def tokenize(self, x_ng: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        tokens = self.feature_ids.expand(x_ng.shape)
        # sort by median-scaled gene values
        sorted_indices = torch.argsort(x_ng, dim=1, descending=True)
        sorted_indices = sorted_indices[:, : self.bert.config.max_position_embeddings]
        ndx = torch.arange(x_ng.shape[0], device=x_ng.device)
        input_ids = tokens[ndx[:, None], sorted_indices]
        # mask out genes with zero expression
        sorted_x_ng = x_ng[ndx[:, None], sorted_indices]
        attention_mask = sorted_x_ng != 0
        # pad genes with zero expression
        input_ids[~attention_mask] = 0
        return input_ids, attention_mask

    def tokenize_with_perturbations(
        self,
        x_ng: torch.Tensor,
        feature_activation: list[str] | None = None,
        feature_deletion: list[str] | None = None,
        feature_map: dict[str, int] | None = None,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        # activation and deletion happen before sorting
        if feature_deletion:
            if not all([g in self.var_names_g for g in feature_deletion]):
                raise ValueError("Some feature_deletion elements are not in self.var_names_g")
            deletion_logic_g = np.isin(self.var_names_g, feature_deletion)
            x_ng[:, deletion_logic_g] = 0
        if feature_activation:
            max_val = x_ng.max()
            for i, g in enumerate(feature_activation[::-1]):
                feature_logic_g = self.var_names_g == g
                if feature_logic_g.sum() != 1:
                    raise ValueError(f"feature_activation element {g} is not in self.var_names_g")
                top_rank_value = max_val + i + 1
                x_ng[:, feature_logic_g] = top_rank_value

        # tokenize and sort to give rank-ordered inputs
        input_ids, attention_mask = self.tokenize(x_ng)

        # feature map is applied after tokenization and sorting
        if feature_map:
            for g, target_token in feature_map.items():
                feature_logic_g = self.var_names_g == g
                if feature_logic_g.sum() != 1:
                    raise ValueError(f"feature_map key {g} not in self.var_names_g")
                initial_token = self.feature_ids[feature_logic_g]
                input_ids[input_ids == initial_token] = target_token

        return input_ids, attention_mask

    def forward(self, x_ng: torch.Tensor, var_names_g: np.ndarray) -> dict[str, torch.Tensor]:
        """
        Args:
            x_ng:
                Gene counts matrix.
            var_names_g:
                The list of the variable names in the input data.
        Returns:
            A dictionary with the loss value.
        """
        assert_columns_and_array_lengths_equal("x_ng", x_ng, "var_names_g", var_names_g)
        assert_arrays_equal("var_names_g", var_names_g, "var_names_g", self.var_names_g)

        input_ids, attention_mask = self.tokenize(x_ng)

        labels = input_ids.clone()
        labels_probs = torch.full(labels.shape, self.mlm_probability, device=x_ng.device)
        labels_probs[~attention_mask] = 0
        masked_indices = torch.bernoulli(labels_probs).bool()
        labels[~masked_indices] = -100  # we only compute loss on masked tokens

        # 80% of the time, we replace masked input tokens with tokenizer.mask_token ([MASK])
        indices_replaced = torch.bernoulli(torch.full(labels.shape, 0.8, device=x_ng.device)).bool() & masked_indices
        input_ids[indices_replaced] = 1  # tokenizer.mask_token_id

        # 10% of the time, we replace masked input tokens with random word
        indices_random = (
            torch.bernoulli(torch.full(labels.shape, 0.5, device=x_ng.device)).bool()
            & masked_indices
            & ~indices_replaced
        )
        random_words = torch.randint(x_ng.shape[1], labels.shape, dtype=torch.long, device=x_ng.device)
        input_ids[indices_random] = random_words[indices_random]

        output = self.bert(
            input_ids=input_ids,
            attention_mask=attention_mask,
            labels=labels,
        )
        return {"loss": output.loss}

    def predict(
        self,
        x_ng: torch.Tensor,
        var_names_g: np.ndarray,
        output_hidden_states: bool = True,
        output_attentions: bool = True,
        output_input_ids: bool = True,
        output_attention_mask: bool = True,
        feature_activation: list[str] | None = None,
        feature_deletion: list[str] | None = None,
        feature_map: dict[str, int] | None = None,
    ) -> dict[str, torch.Tensor | np.ndarray]:
        """
        Send (transformed) data through the model and return outputs.
        Optionally perform in silico perturbations and masking.

        Args:
            x_ng:
                Gene counts matrix.
            var_names_g:
                The list of the variable names in the input data.
            output_hidden_states:
                Whether to return all hidden-states.
            output_attentions:
                Whether to return all attentions.
            output_input_ids:
                Whether to return input ids.
            output_attention_mask:
                Whether to return attention mask.
            feature_activation:
                Specify features whose expression should be set to > max(x_ng) before tokenization (top rank).
            feature_deletion:
                Specify features whose expression should be set to zero before tokenization (remove from inputs).
            feature_map:
                Specify a mapping for input tokens, to be applied before model.

        Returns:
            A dictionary with the inference results.

        .. note::
            In silico perturbations can be achieved in one of three ways:

            1. Use ``feature_map`` to replace a feature token with ``MASK`` (1) or ``PAD`` (0)
               (e.g. ``feature_map={"ENSG0001": 1}`` will replace ``var_names_g`` feature
               ``ENSG0001`` with a ``MASK`` token).
            2. Use ``feature_deletion`` to remove a feature from the cell's inputs, which instead of adding a
               ``PAD`` or ``MASK`` token, will allow another feature to take its place
               (e.g. ``feature_deletion=["ENSG0001"]`` will remove ``var_names_g`` feature ``ENSG0001`` from the input,
               and allow a new feature token to take its place).
            3. Use ``feature_activation`` to move a feature all the way to the top rank position in the input
               (e.g. ``feature_activation=["ENSG0001"]`` will make ``var_names_g`` feature ``ENSG0001`` the first in
               rank order. Multiple input features will be ranked according to their order in the input list).

            Number (2) and (3) are described in the Geneformer paper under "In silico perturbation" in the
            Methods section.
        """
        assert_columns_and_array_lengths_equal("x_ng", x_ng, "var_names_g", var_names_g)
        assert_arrays_equal("var_names_g", var_names_g, "var_names_g", self.var_names_g)

        input_ids, attention_mask = self.tokenize_with_perturbations(
            x_ng, feature_activation=feature_activation, feature_deletion=feature_deletion, feature_map=feature_map
        )

        output = self.bert(
            input_ids=input_ids,
            attention_mask=attention_mask,
            output_hidden_states=output_hidden_states,
            output_attentions=output_attentions,
        )
        if output_input_ids:
            output["input_ids"] = input_ids
        if output_attention_mask:
            output["attention_mask"] = attention_mask
        return output



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import copy
import math
from collections.abc import Callable
from typing import Any

import torch

from cellarium.ml.utilities.testing import assert_nonnegative, assert_positive


def copy_module(
    module: torch.nn.Module, self_device: torch.device, copy_device: torch.device
) -> tuple[torch.nn.Module, torch.nn.Module]:
    """
    Return an original module on ``self_device`` and its copy on ``copy_device``.
    If the module is on meta device then it is moved to ``self_device`` efficiently using ``to_empty`` method.

    Args:
        module:
            The module to copy.
        self_device:
            The device to send the original module to.
        copy_device:
            The device to copy the module to.

    Returns:
        A tuple of the original module and its copy.
    """
    module_copy = copy.deepcopy(module)
    if any(param.device.type == "meta" for param in module.parameters()) or any(
        buffer.device.type == "meta" for buffer in module.buffers()
    ):
        module.to_empty(device=self_device)
        module_copy.to_empty(device=copy_device)
    else:
        module.to(device=self_device)
        module_copy.to(device=copy_device)
    return module, module_copy


def train_val_split(n_samples: int, train_size: float | int | None, val_size: float | int | None) -> tuple[int, int]:
    """
    Validate the train and validation sizes and return the number of samples for each.

    Args:
        n_samples:
            The number of samples in the dataset.
        train_size:
            Size of the train split. If :class:`float`, should be between ``0.0`` and ``1.0`` and represent
            the proportion of the dataset to include in the train split. If :class:`int`, represents
            the absolute number of train samples. If ``None``, the value is automatically set to the complement
            of the ``val_size``.
        val_size:
            Size of the validation split. If :class:`float`, should be between ``0.0`` and ``1.0`` and represent
            the proportion of the dataset to include in the validation split. If :class:`int`, represents
            the absolute number of validation samples. If ``None``, the value is set to the complement of
            the ``train_size``. If ``train_size`` is also ``None``, it will be set to ``0``.

    Returns:
        A tuple of the number of samples for training and validation.
    """
    if train_size is None and val_size is None:
        n_val = 0
        n_train = n_samples

    elif train_size is None:
        if isinstance(val_size, int):
            assert_nonnegative("val_size", val_size)
            n_val = val_size
        elif isinstance(val_size, float):
            if val_size < 0.0 or val_size >= 1.0:
                raise ValueError(f"If validation size is a float it should be 0.0 <= val_size < 1.0. Got {val_size}")
            n_val = math.ceil(n_samples * val_size)
        n_train = n_samples - n_val

    elif val_size is None:
        if isinstance(train_size, int):
            assert_positive("train_size", train_size)
            n_train = train_size
        elif isinstance(train_size, float):
            if train_size <= 0.0 or train_size > 1.0:
                raise ValueError(f"If train size is a float it should be 0.0 < train_size <= 1.0. Got {train_size}")
            n_train = math.ceil(n_samples * train_size)
        n_val = n_samples - n_train

    else:
        if isinstance(train_size, int):
            n_train = train_size
        elif isinstance(train_size, float):
            n_train = math.ceil(n_samples * train_size)

        if isinstance(val_size, int):
            n_val = val_size
        elif isinstance(val_size, float):
            n_val = math.ceil(n_samples * val_size)

    if n_train + n_val > n_samples:
        raise ValueError(
            f"Size of train and validation splits ({n_train + n_val}) is greater than "
            f"the number of samples ({n_samples})"
        )

    return n_train, n_val


def call_func_with_batch(
    func: Callable,
    batch: dict[str, Any],
) -> Any:
    """
    Call a function with a batch dictionary. If the function is a method of a :class:`CellariumModule`, the function
    is called with the batch dictionary as its only argument. Otherwise, the function is called with the keys from
    the batch dictionary that are present in its annotations. If the function has a ``kwargs`` annotation, all keys from
    the batch dictionary are passed to the function.

    Args:
        func:
            The function to call.
        batch:
            The batch dictionary.
    """
    from cellarium.ml import CellariumModule

    if isinstance(getattr(func, "__self__", None), CellariumModule):
        # in case module is a CellariumModule, e.g. PCA checkpoint is used as a transform
        return func(batch)

    ann = func.__annotations__
    input_keys = {key for key in ann if key != "return" and key in batch}
    # allow all keys to be passed to the function
    if "kwargs" in ann:
        input_keys |= batch.keys()
    return func(**{key: batch[key] for key in input_keys})



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import torch
import torch.distributed as dist


class GatherLayer(torch.autograd.Function):
    """Gather tensors from all process, supporting backward propagation."""

    @staticmethod
    def forward(ctx, input: torch.Tensor) -> tuple[torch.Tensor, ...]:  # type: ignore
        output = [torch.empty_like(input) for _ in range(dist.get_world_size())]
        dist.all_gather(output, input)
        return tuple(output)

    @staticmethod
    def backward(ctx, *grads: torch.Tensor) -> torch.Tensor:
        all_grads = torch.stack(grads)
        dist.all_reduce(all_grads, op=dist.ReduceOp.SUM)
        return all_grads[dist.get_rank()]



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import warnings
from collections.abc import Iterable
from typing import Any

import lightning.pytorch as pl
import numpy as np
import torch
from lightning.pytorch.utilities.types import STEP_OUTPUT, OptimizerLRSchedulerConfig

from cellarium.ml.core.pipeline import CellariumPipeline
from cellarium.ml.models import CellariumModel
from cellarium.ml.utilities.core import copy_module


class CellariumModule(pl.LightningModule):
    """
    ``CellariumModule`` organizes code into following sections:

        * :attr:`transforms`: A list of transforms to apply to the input data before passing it to the model.
        * :attr:`model`: A :class:`cellarium.ml.models.CellariumModel` to train with
          :meth:`training_step` method and epoch end hooks.
        * :attr:`optim_fn` and :attr:`optim_kwargs`: A Pytorch optimizer class and its keyword arguments.
        * :attr:`scheduler_fn` and :attr:`scheduler_kwargs`: A Pytorch lr scheduler class and its
          keyword arguments.

    Args:
        transforms:
            A list of transforms to apply to the input data before passing it to the model.
            If ``None``, no transforms are applied.
        model:
            A :class:`cellarium.ml.models.CellariumModel` to train.
        optim_fn:
            A Pytorch optimizer class, e.g., :class:`~torch.optim.Adam`. If ``None``,
            no optimizer is used.
        optim_kwargs:
            Keyword arguments for optimiser.
        scheduler_fn:
            A Pytorch lr scheduler class, e.g., :class:`~torch.optim.lr_scheduler.CosineAnnealingLR`.
        scheduler_kwargs:
            Keyword arguments for lr scheduler.
        is_initialized:
            Whether the model has been initialized. This is set to ``False`` by default under the assumption that
            ``torch.device("meta")`` context was used and is set to ``True`` after
            the first call to :meth:`configure_model`.
    """

    def __init__(
        self,
        transforms: Iterable[torch.nn.Module] | None = None,
        model: CellariumModel | None = None,
        optim_fn: type[torch.optim.Optimizer] | None = None,
        optim_kwargs: dict[str, Any] | None = None,
        scheduler_fn: type[torch.optim.lr_scheduler.LRScheduler] | None = None,
        scheduler_kwargs: dict[str, Any] | None = None,
        is_initialized: bool = False,
    ) -> None:
        super().__init__()
        self.save_hyperparameters(logger=False)
        self.pipeline: CellariumPipeline | None = None

    def configure_model(self) -> None:
        # This hook is called during each of fit/val/test/predict stages in the same process, so ensure that
        # implementation of this hook is idempotent, i.e., after the first time the hook is called, subsequent
        # calls to it should be a no-op.
        if self.pipeline is not None:
            return

        # Steps involved in configuring the model:
        # 1. Make a copy of modules on the meta device and assign to hparams.
        # 2. Send the original modules to the host device and add to self.pipeline.
        # 3. Reset the model parameters if it has not been initialized before.
        # For more context, see discussions in
        # https://dev-discuss.pytorch.org/t/state-of-model-creation-initialization-seralization-in-pytorch-core/1240
        #
        # Benefits of this approach:
        # 1. The checkpoint stores modules on the meta device.
        # 2. Loading from a checkpoint skips a wasteful step of initializing module parameters
        #    before loading the state_dict.
        # 3. The module parameters are directly initialized on the host gpu device instead of being initialized
        #    on the cpu and then moved to the gpu device (given that modules were instantiated under
        #    the ``torch.device("meta")`` context).
        model, self.hparams["model"] = copy_module(
            self.hparams["model"], self_device=self.device, copy_device=torch.device("meta")
        )
        if self.hparams["transforms"]:
            transforms, self.hparams["transforms"] = zip(
                *(
                    copy_module(transform, self_device=self.device, copy_device=torch.device("meta"))
                    for transform in self.hparams["transforms"]
                )
            )
        else:
            transforms = None

        self.pipeline = CellariumPipeline(transforms)
        if model is None:
            raise ValueError(f"`model` must be an instance of {CellariumModel}. Got {model}")
        self.pipeline.append(model)

        if not self.hparams["is_initialized"]:
            model.reset_parameters()
            self.hparams["is_initialized"] = True

    @property
    def model(self) -> CellariumModel:
        """The model"""
        if self.pipeline is None:
            raise RuntimeError("The model is not configured. Call `configure_model` before accessing the model.")

        return self.pipeline[-1]

    @property
    def transforms(self) -> CellariumPipeline:
        """The transforms pipeline"""
        if self.pipeline is None:
            raise RuntimeError("The model is not configured. Call `configure_model` before accessing the model.")

        return self.pipeline[:-1]

    def training_step(  # type: ignore[override]
        self, batch: dict[str, np.ndarray | torch.Tensor], batch_idx: int
    ) -> torch.Tensor | None:
        """
        Forward pass for training step.

        Args:
            batch:
                A dictionary containing the batch data.
            batch_idx:
                The index of the batch.

        Returns:
            Loss tensor or ``None`` if no loss.
        """
        if self.pipeline is None:
            raise RuntimeError("The model is not configured. Call `configure_model` before accessing the model.")

        output = self.pipeline(batch)
        loss = output.get("loss")
        if loss is not None:
            # Logging to TensorBoard by default
            self.log("train_loss", loss)
        return loss

    def forward(self, batch: dict[str, np.ndarray | torch.Tensor]) -> dict[str, np.ndarray | torch.Tensor]:
        """
        Forward pass for inference step.

        Args:
            batch: A dictionary containing the batch data.

        Returns:
            A dictionary containing the batch data and inference outputs.
        """
        if self.pipeline is None:
            raise RuntimeError("The model is not configured. Call `configure_model` before accessing the model.")

        return self.pipeline.predict(batch)

    def validation_step(self, batch: dict[str, Any], batch_idx: int) -> None:
        """
        Forward pass for validation step.

        Args:
            batch:
                A dictionary containing the batch data.
            batch_idx:
                The index of the batch.

        Returns:
            None
        """
        if self.pipeline is None:
            raise RuntimeError("The model is not configured. Call `configure_model` before accessing the model.")

        self.pipeline.validate(batch)

    def configure_optimizers(self) -> OptimizerLRSchedulerConfig | None:
        """Configure optimizers for the model."""
        optim_fn = self.hparams["optim_fn"]
        optim_kwargs = self.hparams["optim_kwargs"] or {}
        scheduler_fn = self.hparams["scheduler_fn"]
        scheduler_kwargs = self.hparams["scheduler_kwargs"] or {}

        if optim_fn is None:
            if optim_kwargs:
                warnings.warn("Optimizer kwargs are provided but no optimizer is defined.", UserWarning)
            if scheduler_fn is not None:
                warnings.warn("Scheduler is defined but no optimizer is defined.", UserWarning)
            return None

        optim_config: OptimizerLRSchedulerConfig = {"optimizer": optim_fn(self.model.parameters(), **optim_kwargs)}
        if scheduler_fn is not None:
            scheduler = scheduler_fn(optim_config["optimizer"], **scheduler_kwargs)
            optim_config["lr_scheduler"] = {"scheduler": scheduler, "interval": "step"}
        return optim_config

    def on_train_epoch_start(self) -> None:
        """
        Calls the ``set_epoch`` method on the iterable dataset of the given dataloader.

        If the dataset is ``IterableDataset`` and has ``set_epoch`` method defined, then
        ``set_epoch`` must be called at the beginning of every epoch to ensure shuffling
        applies a new ordering. This has no effect if shuffling is off.
        """
        # dataloader is wrapped in a combined loader and can be accessed via
        # flattened property which returns a list of dataloaders
        # https://lightning.ai/docs/pytorch/stable/api/lightning.pytorch.utilities.combined_loader.html
        combined_loader = self.trainer.fit_loop._combined_loader
        assert combined_loader is not None
        dataloaders = combined_loader.flattened
        for dataloader in dataloaders:
            dataset = dataloader.dataset
            set_epoch = getattr(dataset, "set_epoch", None)
            if callable(set_epoch):
                set_epoch(self.current_epoch)

    def on_train_start(self) -> None:
        """
        Calls the ``on_train_start`` method on the :attr:`model` attribute.
        If the :attr:`model` attribute has ``on_train_start`` method defined, then
        ``on_train_start`` must be called at the beginning of training.
        """
        on_train_start = getattr(self.model, "on_train_start", None)
        if callable(on_train_start):
            on_train_start(self.trainer)

    def on_train_epoch_end(self) -> None:
        """
        Calls the ``on_epoch_end`` method on the :attr:`model` attribute.
        If the :attr:`model` attribute has ``on_epoch_end`` method defined, then
        ``on_epoch_end`` must be called at the end of every epoch.
        """
        on_epoch_end = getattr(self.model, "on_epoch_end", None)
        if callable(on_epoch_end):
            on_epoch_end(self.trainer)

    def on_train_batch_end(self, outputs: STEP_OUTPUT, batch: Any, batch_idx: int) -> None:
        """
        Calls the ``on_batch_end`` method on the module.
        """
        on_batch_end = getattr(self.model, "on_batch_end", None)
        if callable(on_batch_end):
            on_batch_end(self.trainer)



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Sequence
from functools import cache
from typing import Any

import numpy as np
import torch
from torch import nn

from cellarium.ml.utilities.testing import (
    assert_columns_and_array_lengths_equal,
)


class Filter(nn.Module):
    """
    Filter gene counts by a list of features.

    .. math::

        \\mathrm{mask}_g = \\mathrm{feature}_g \\in \\mathrm{filter\\_list}

        y_{ng} = x_{ng}[:, \\mathrm{mask}_g]

    Args:
        filter_list: A list of features to filter by.
    """

    def __init__(self, filter_list: Sequence[str]) -> None:
        super().__init__()
        self.filter_list = np.array(filter_list)
        if len(self.filter_list) == 0:
            raise ValueError(f"`filter_list` must not be empty. Got {self.filter_list}")

    @cache
    def filter(self, var_names_g: tuple) -> np.ndarray[Any, np.dtype[np.int_]]:
        """
        Args:
            var_names_g: The list of the variable names in the input data.

        Returns:
            An array of indices of the features in ``var_names_g`` that are in :attr:`filter_list`.
        """
        mask = np.isin(var_names_g, self.filter_list)
        if not np.any(mask):
            raise AssertionError("No features in `var_names_g` matched the `filter_list`")
        mask_indices = np.where(mask)[0]
        return mask_indices

    def forward(self, x_ng: torch.Tensor, var_names_g: np.ndarray) -> dict[str, torch.Tensor | np.ndarray]:
        """
        .. note::

            When used with :class:`~cellarium.ml.core.CellariumModule` or :class:`~cellarium.ml.core.CellariumPipeline`,
            ``x_ng`` and ``var_names_g`` keys in the input dictionary will be overwritten with the filtered values.

        Args:
            x_ng:
                Gene counts.
            var_names_g:
                The list of the variable names in the input data.

        Returns:
            A dictionary with the following keys:

            - ``x_ng``: Gene counts filtered by :attr:`filter_list`.
            - ``var_names_g``: The list of the variable names in the input data filtered by :attr:`filter_list`.
        """
        assert_columns_and_array_lengths_equal("x_ng", x_ng, "var_names_g", var_names_g)

        filter_indices = self.filter(tuple(var_names_g.tolist()))
        ndx = torch.arange(x_ng.shape[0])
        x_ng = x_ng[ndx[:, None], filter_indices]
        var_names_g = var_names_g[filter_indices]

        return {"x_ng": x_ng, "var_names_g": var_names_g}

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(filter_list={self.filter_list})"



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from pathlib import Path

import lightning.pytorch as pl
import pytest

from cellarium.ml import CellariumAnnDataDataModule, CellariumModule
from cellarium.ml.data import DistributedAnnDataCollection
from cellarium.ml.utilities.core import train_val_split
from cellarium.ml.utilities.data import AnnDataField, densify
from tests.common import BoringModel


@pytest.mark.parametrize("batch_size", [50, None])
def test_datamodule(tmp_path: Path, batch_size: int | None) -> None:
    datamodule = CellariumAnnDataDataModule(
        DistributedAnnDataCollection(
            filenames="https://storage.googleapis.com/dsp-cellarium-cas-public/test-data/test_0.h5ad",
            shard_size=100,
        ),
        batch_size=100,
        batch_keys={
            "x_ng": AnnDataField(attr="X", convert_fn=densify),
        },
    )
    module = CellariumModule(model=BoringModel())
    trainer = pl.Trainer(accelerator="cpu", devices=1, max_steps=1, default_root_dir=tmp_path)
    trainer.fit(module, datamodule)

    ckpt_path = str(tmp_path / "lightning_logs/version_0/checkpoints/epoch=0-step=1.ckpt")
    adata = datamodule.dadc.adatas[0].adata
    kwargs = {"dadc": adata}
    if batch_size is not None:
        kwargs["batch_size"] = batch_size
    loaded_datamodule = CellariumAnnDataDataModule.load_from_checkpoint(ckpt_path, **kwargs)

    assert loaded_datamodule.batch_keys == datamodule.batch_keys
    assert loaded_datamodule.batch_size == batch_size or datamodule.batch_size
    assert loaded_datamodule.dadc is adata


@pytest.mark.parametrize(
    "n_samples, train_size, val_size, n_train_expected, n_val_expected",
    [
        (100, 0.8, 0.1, 80, 10),
        (100, 0.8, 0.2, 80, 20),
        (100, 80, 0.2, 80, 20),
        (100, 80, 20, 80, 20),
        (100, 80, None, 80, 20),
        (100, None, 20, 80, 20),
        (100, 0.8, None, 80, 20),
        (100, None, 0.2, 80, 20),
        (100, None, None, 100, 0),
    ],
)
def test_train_val_split(
    n_samples: int,
    train_size: float | int | None,
    val_size: float | int | None,
    n_train_expected: int,
    n_val_expected: int,
) -> None:
    n_train_actual, n_val_actual = train_val_split(n_samples, train_size, val_size)
    assert n_train_actual == n_train_expected
    assert n_val_actual == n_val_expected



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause


from collections.abc import Sequence
from typing import Literal

import numpy as np
import pyro
import pyro.distributions as dist
import torch
from pyro.nn.module import PyroParam, _unconstrain
from torch.distributions import constraints

from cellarium.ml.models.model import CellariumModel, PredictMixin
from cellarium.ml.utilities.testing import (
    assert_arrays_equal,
    assert_columns_and_array_lengths_equal,
)


class ProbabilisticPCA(CellariumModel, PredictMixin):
    """
    Probabilistic PCA implemented in Pyro.

    Two flavors of probabilistic PCA are available - marginalized pPCA [1]
    and linear VAE [2].

    **References:**

    1. `Probabilistic Principal Component Analysis (Tipping et al.)
       <https://www.robots.ox.ac.uk/~cvrg/hilary2006/ppca.pdf>`_.
    2. `Understanding Posterior Collapse in Generative Latent Variable Models (Lucas et al.)
       <https://openreview.net/pdf?id=r1xaVLUYuE>`_.

    Args:
        n_obs:
            Number of cells.
        var_names_g:
            The variable names schema for the input data validation.
        n_components:
            Number of principal components.
        ppca_flavor:
            Type of the PPCA model. Has to be one of `marginalized` or `linear_vae`.
        mean_g:
            Mean gene expression of the input data. If ``None`` then the mean is
            set to a learnable parameter.
        W_init_scale:
            Scale of the random initialization of the `W_kg` parameter.
        sigma_init_scale:
            Initialization value of the `sigma` parameter.
        seed:
            Random seed used to initialize parameters.
        elbo:
            ELBO loss function. Should be a subclass of :class:`~pyro.infer.ELBO`.
            If ``None``, defaults to :class:`~pyro.infer.Trace_ELBO`.
    """

    def __init__(
        self,
        n_obs: int,
        var_names_g: Sequence[str],
        n_components: int,
        ppca_flavor: Literal["marginalized", "linear_vae"],
        mean_g: torch.Tensor | None = None,
        W_init_scale: float = 1.0,
        sigma_init_scale: float = 1.0,
        seed: int = 0,
        elbo: pyro.infer.ELBO | None = None,
    ):
        super().__init__()

        self.n_obs = n_obs
        self.var_names_g = np.array(var_names_g)
        n_vars = len(self.var_names_g)
        self.n_vars = n_vars
        self.n_components = n_components
        self.ppca_flavor = ppca_flavor
        self.elbo = elbo or pyro.infer.Trace_ELBO()

        if isinstance(mean_g, torch.Tensor) and mean_g.dim():
            assert mean_g.shape == (n_vars,), f"Expected meang_g to have a shape ({n_vars},) but found {mean_g.shape}."
        if mean_g is None:
            # make mean_g a learnable parameter
            self.mean_g = torch.nn.Parameter(torch.empty(n_vars))
        else:
            self.register_buffer("mean_g", mean_g)

        self.seed = seed
        # model parameters
        self.W_init_scale = W_init_scale
        self.sigma_init_scale = sigma_init_scale
        self.W_kg = torch.nn.Parameter(torch.empty(n_components, n_vars))
        self.sigma = PyroParam(torch.empty(()), constraint=constraints.positive)  # type: ignore[call-arg]
        self.reset_parameters()

    def reset_parameters(self) -> None:
        rng_device = self.W_kg.device.type if self.W_kg.device.type != "meta" else "cpu"
        rng = torch.Generator(device=rng_device)
        rng.manual_seed(self.seed)
        if isinstance(self.mean_g, torch.nn.Parameter):
            self.mean_g.data.zero_()
        self.W_kg.data.normal_(0, self.W_init_scale, generator=rng)
        self.sigma_unconstrained.data.fill_(_unconstrain(torch.as_tensor(self.sigma_init_scale), constraints.positive))

    def forward(self, x_ng: torch.Tensor, var_names_g: np.ndarray) -> dict[str, torch.Tensor | None]:
        """
        Args:
            x_ng:
                Gene counts matrix.
            var_names_g:
                The list of the variable names in the input data.

        Returns:
            A dictionary with the loss value.
        """
        assert_columns_and_array_lengths_equal("x_ng", x_ng, "var_names_g", var_names_g)
        assert_arrays_equal("var_names_g", var_names_g, "var_names_g", self.var_names_g)

        loss = self.elbo.differentiable_loss(self.model, self.guide, x_ng)  # type: ignore[attr-defined]
        return {"loss": loss}

    def model(self, x_ng: torch.Tensor) -> None:
        with pyro.plate("cells", size=self.n_obs, subsample_size=x_ng.shape[0]):
            if self.ppca_flavor == "marginalized":
                pyro.sample(
                    "counts",
                    dist.LowRankMultivariateNormal(  # type: ignore[attr-defined]
                        loc=self.mean_g,
                        cov_factor=self.W_kg.T,  # type: ignore[attr-defined]
                        cov_diag=self.sigma**2 * x_ng.new_ones(self.n_vars),  # type: ignore[operator]
                    ),
                    obs=x_ng,
                )
            else:
                z_nk = pyro.sample(
                    "z",
                    dist.Normal(x_ng.new_zeros(self.n_components), 1).to_event(1),  # type: ignore[attr-defined]
                )
                pyro.sample(
                    "counts",
                    dist.Normal(self.mean_g + z_nk @ self.W_kg, self.sigma).to_event(1),  # type: ignore[attr-defined]
                    obs=x_ng,
                )

    def guide(self, x_ng: torch.Tensor) -> None:
        if self.ppca_flavor == "marginalized":
            return

        with pyro.plate("cells", size=self.n_obs, subsample_size=x_ng.shape[0]):
            V_gk = torch.linalg.solve(self.M_kk, self.W_kg).T
            D_k = self.sigma / torch.sqrt(torch.diag(self.M_kk))
            pyro.sample("z", dist.Normal((x_ng - self.mean_g) @ V_gk, D_k).to_event(1))  # type: ignore[attr-defined]

    def predict(self, x_ng: torch.Tensor, var_names_g: np.ndarray) -> dict[str, np.ndarray | torch.Tensor]:
        """
        Centering and embedding of the input data ``x_ng`` into the principal component space.

        .. note::
           Gradients are disabled, used for inference only.

        Args:
            x_ng:
                Gene counts matrix.
            var_names_g:
                The list of the variable names in the input data.

        Returns:
            A dictionary with the following keys:

            - ``z_nk``: Embedding of the input data into the principal component space.
        """
        assert_columns_and_array_lengths_equal("x_ng", x_ng, "var_names_g", var_names_g)
        assert_arrays_equal("var_names_g", var_names_g, "var_names_g", self.var_names_g)

        V_gk = torch.linalg.solve(self.M_kk, self.W_kg).T
        z_nk = (x_ng - self.mean_g) @ V_gk
        return {"z_nk": z_nk}

    @property
    def M_kk(self) -> torch.Tensor:
        return self.W_kg @ self.W_kg.T + self.sigma**2 * torch.eye(self.n_components, device=self.sigma.device)  # type: ignore[attr-defined, operator]

    @property
    @torch.inference_mode()
    def L_k(self) -> torch.Tensor:
        r"""
        Vector with elements given by the PC eigenvalues.

        .. note::
           Gradients are disabled, used for inference only.
        """
        S_k = torch.linalg.svdvals(self.W_kg.T)  # type: ignore[attr-defined]
        return S_k**2 + self.sigma**2  # type: ignore[operator]

    @property
    @torch.inference_mode()
    def U_gk(self) -> torch.Tensor:
        r"""
        Principal components corresponding to eigenvalues ``L_k``.

        .. note::
           Gradients are disabled, used for inference only.
        """
        return torch.linalg.svd(self.W_kg.T, full_matrices=False).U  # type: ignore[attr-defined]

    @property
    @torch.inference_mode()
    def W_variance(self) -> float:
        r"""
        .. note::
           Gradients are disabled, used for inference only.
        """
        return torch.trace(self.W_kg.T @ self.W_kg).item()  # type: ignore[attr-defined]

    @property
    @torch.inference_mode()
    def sigma_variance(self) -> float:
        r"""
        .. note::
           Gradients are disabled, used for inference only.
        """
        return (self.n_vars * self.sigma**2).item()  # type: ignore[operator]



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import os
from pathlib import Path
from typing import Any

import pytest

from cellarium.ml.cli import main

devices = os.environ.get("TEST_DEVICES", "1")

CONFIGS = [
    {
        "model_name": "geneformer",
        "subcommand": "fit",
        "fit": {
            "model": {
                "model": {
                    "class_path": "cellarium.ml.models.Geneformer",
                    "init_args": {
                        "hidden_size": "2",
                        "num_hidden_layers": "1",
                        "num_attention_heads": "1",
                        "intermediate_size": "4",
                        "max_position_embeddings": "2",
                    },
                },
                "optim_fn": "torch.optim.Adam",
            },
            "data": {
                "dadc": {
                    "class_path": "cellarium.ml.data.DistributedAnnDataCollection",
                    "init_args": {
                        "filenames": "https://storage.googleapis.com/dsp-cellarium-cas-public/test-data/test_0.h5ad",
                        "shard_size": "100",
                        "max_cache_size": "2",
                        "obs_columns_to_validate": [],
                    },
                },
                "batch_keys": {
                    "x_ng": {
                        "attr": "X",
                        "convert_fn": "cellarium.ml.utilities.data.densify",
                    },
                    "var_names_g": {"attr": "var_names"},
                },
                "batch_size": "5",
                "num_workers": "1",
            },
            "trainer": {
                "accelerator": "cpu",
                "devices": devices,
                "max_steps": "1",
            },
        },
    },
    {
        "model_name": "geneformer",
        "subcommand": "predict",
        "predict": {
            "model": {
                "model": {
                    "class_path": "cellarium.ml.models.Geneformer",
                    "init_args": {
                        "hidden_size": "2",
                        "num_hidden_layers": "1",
                        "num_attention_heads": "1",
                        "intermediate_size": "4",
                        "max_position_embeddings": "2",
                    },
                },
            },
            "data": {
                "dadc": {
                    "class_path": "cellarium.ml.data.DistributedAnnDataCollection",
                    "init_args": {
                        "filenames": "https://storage.googleapis.com/dsp-cellarium-cas-public/test-data/test_0.h5ad",
                        "shard_size": "100",
                        "max_cache_size": "2",
                        "obs_columns_to_validate": [],
                    },
                },
                "batch_keys": {
                    "x_ng": {
                        "attr": "X",
                        "convert_fn": "cellarium.ml.utilities.data.densify",
                    },
                    "var_names_g": {"attr": "var_names"},
                },
                "batch_size": "5",
                "num_workers": "1",
            },
            "trainer": {
                "accelerator": "cpu",
                "devices": devices,
                "max_steps": "1",
                "limit_predict_batches": "1",
            },
            "return_predictions": "false",
        },
    },
    {
        "model_name": "probabilistic_pca",
        "subcommand": "fit",
        "fit": {
            "model": {
                "model": {
                    "class_path": "cellarium.ml.models.ProbabilisticPCA",
                    "init_args": {
                        "n_components": "2",
                        "ppca_flavor": "marginalized",
                    },
                },
                "optim_fn": "torch.optim.Adam",
            },
            "data": {
                "dadc": {
                    "class_path": "cellarium.ml.data.DistributedAnnDataCollection",
                    "init_args": {
                        "filenames": "https://storage.googleapis.com/dsp-cellarium-cas-public/test-data/test_{0..1}.h5ad",
                        "shard_size": "100",
                        "max_cache_size": "2",
                        "obs_columns_to_validate": [],
                    },
                },
                "batch_keys": {
                    "x_ng": {
                        "attr": "X",
                        "convert_fn": "cellarium.ml.utilities.data.densify",
                    },
                    "var_names_g": {"attr": "var_names"},
                },
                "batch_size": "50",
                "shuffle": "true",
                "num_workers": "2",
            },
            "trainer": {
                "accelerator": "cpu",
                "devices": devices,
                "max_steps": "4",
            },
        },
    },
    {
        "model_name": "onepass_mean_var_std",
        "subcommand": "fit",
        "fit": {
            "model": {
                "transforms": [
                    {
                        "class_path": "cellarium.ml.transforms.NormalizeTotal",
                        "init_args": {"target_count": "10_000"},
                    },
                    "cellarium.ml.transforms.Log1p",
                ],
                "model": "cellarium.ml.models.OnePassMeanVarStd",
            },
            "data": {
                "dadc": {
                    "class_path": "cellarium.ml.data.DistributedAnnDataCollection",
                    "init_args": {
                        "filenames": "https://storage.googleapis.com/dsp-cellarium-cas-public/test-data/test_0.h5ad",
                        "shard_size": "100",
                        "max_cache_size": "2",
                        "obs_columns_to_validate": ["total_mrna_umis"],
                    },
                },
                "batch_keys": {
                    "x_ng": {
                        "attr": "X",
                        "convert_fn": "cellarium.ml.utilities.data.densify",
                    },
                    "var_names_g": {"attr": "var_names"},
                    "total_mrna_umis_n": {
                        "attr": "obs",
                        "key": "total_mrna_umis",
                    },
                },
                "batch_size": "50",
                "num_workers": "2",
            },
            "trainer": {
                "accelerator": "cpu",
                "devices": devices,
            },
        },
    },
    {
        "model_name": "incremental_pca",
        "subcommand": "fit",
        "fit": {
            "model": {
                "transforms": [
                    {
                        "class_path": "cellarium.ml.transforms.NormalizeTotal",
                        "init_args": {"target_count": "10_000"},
                    },
                    "cellarium.ml.transforms.Log1p",
                ],
                "model": {
                    "class_path": "cellarium.ml.models.IncrementalPCA",
                    "init_args": {"n_components": "50"},
                },
            },
            "data": {
                "dadc": {
                    "class_path": "cellarium.ml.data.DistributedAnnDataCollection",
                    "init_args": {
                        "filenames": "https://storage.googleapis.com/dsp-cellarium-cas-public/test-data/test_{0..1}.h5ad",
                        "shard_size": "100",
                        "max_cache_size": "2",
                        "obs_columns_to_validate": ["total_mrna_umis"],
                    },
                },
                "batch_keys": {
                    "x_ng": {
                        "attr": "X",
                        "convert_fn": "cellarium.ml.utilities.data.densify",
                    },
                    "var_names_g": {"attr": "var_names"},
                    "total_mrna_umis_n": {
                        "attr": "obs",
                        "key": "total_mrna_umis",
                    },
                },
                "batch_size": "50",
                "num_workers": "2",
            },
            "trainer": {
                "accelerator": "cpu",
                "devices": devices,
            },
        },
    },
    {
        "model_name": "incremental_pca",
        "subcommand": "predict",
        "predict": {
            "model": {
                "transforms": [
                    {
                        "class_path": "cellarium.ml.transforms.NormalizeTotal",
                        "init_args": {"target_count": "10_000"},
                    },
                    "cellarium.ml.transforms.Log1p",
                ],
                "model": {
                    "class_path": "cellarium.ml.models.IncrementalPCA",
                    "init_args": {"n_components": "50"},
                },
            },
            "data": {
                "dadc": {
                    "class_path": "cellarium.ml.data.DistributedAnnDataCollection",
                    "init_args": {
                        "filenames": "https://storage.googleapis.com/dsp-cellarium-cas-public/test-data/test_{0..1}.h5ad",
                        "shard_size": "100",
                        "max_cache_size": "2",
                        "obs_columns_to_validate": ["total_mrna_umis"],
                    },
                },
                "batch_keys": {
                    "x_ng": {
                        "attr": "X",
                        "convert_fn": "cellarium.ml.utilities.data.densify",
                    },
                    "var_names_g": {"attr": "var_names"},
                    "obs_names_n": {"attr": "obs_names"},
                    "total_mrna_umis_n": {
                        "attr": "obs",
                        "key": "total_mrna_umis",
                    },
                },
                "batch_size": "50",
                "num_workers": "2",
            },
            "trainer": {
                "accelerator": "cpu",
                "devices": devices,
                "callbacks": {
                    "class_path": "cellarium.ml.callbacks.PredictionWriter",
                    "init_args": {"output_dir": "./output"},
                },
            },
            "return_predictions": "false",
        },
    },
    {
        "model_name": "logistic_regression",
        "subcommand": "fit",
        "fit": {
            "model": {
                "transforms": [
                    {
                        "class_path": "cellarium.ml.transforms.Filter",
                        "init_args": {
                            "filter_list": [
                                "ENSG00000187642",
                                "ENSG00000078808",
                                "ENSG00000272106",
                                "ENSG00000162585",
                                "ENSG00000272088",
                                "ENSG00000204624",
                                "ENSG00000162490",
                                "ENSG00000177000",
                                "ENSG00000011021",
                            ]
                        },
                    }
                ],
                "model": "cellarium.ml.models.LogisticRegression",
                "optim_fn": "torch.optim.Adam",
            },
            "data": {
                "dadc": {
                    "class_path": "cellarium.ml.data.DistributedAnnDataCollection",
                    "init_args": {
                        "filenames": "https://storage.googleapis.com/dsp-cellarium-cas-public/test-data/test_{0..1}.h5ad",
                        "shard_size": "100",
                        "max_cache_size": "2",
                        "obs_columns_to_validate": ["cell_type"],
                    },
                },
                "batch_keys": {
                    "x_ng": {
                        "attr": "X",
                        "convert_fn": "cellarium.ml.utilities.data.densify",
                    },
                    "var_names_g": {
                        "attr": "var_names",
                    },
                    "y_n": {
                        "attr": "obs",
                        "key": "cell_type",
                        "convert_fn": "cellarium.ml.utilities.data.categories_to_codes",
                    },
                },
                "batch_size": "50",
                "shuffle": "true",
                "num_workers": "2",
            },
            "trainer": {
                "accelerator": "cpu",
                "devices": devices,
                "max_steps": "4",
            },
        },
    },
    {
        "model_name": "tdigest",
        "subcommand": "fit",
        "fit": {
            "model": {"model": "cellarium.ml.models.TDigest"},
            "data": {
                "dadc": {
                    "class_path": "cellarium.ml.data.DistributedAnnDataCollection",
                    "init_args": {
                        "filenames": "https://storage.googleapis.com/dsp-cellarium-cas-public/test-data/test_{0..1}.h5ad",
                        "shard_size": "100",
                        "max_cache_size": "2",
                        "obs_columns_to_validate": [],
                    },
                },
                "batch_keys": {
                    "x_ng": {
                        "attr": "X",
                        "convert_fn": "cellarium.ml.utilities.data.densify",
                    },
                    "var_names_g": {"attr": "var_names"},
                },
                "batch_size": "50",
                "num_workers": "2",
            },
            "trainer": {
                "accelerator": "cpu",
                "devices": devices,
            },
        },
    },
]


@pytest.mark.parametrize("config", CONFIGS)
def test_cpu_multi_device(config: dict[str, Any]):
    if config["subcommand"] == "predict":
        assert config["predict"]["return_predictions"] == "false"
    main(config)


def test_checkpoint_loader(tmp_path: Path) -> None:
    onepass_config = f"""
    model:
      transforms:
        - class_path: cellarium.ml.transforms.NormalizeTotal
          init_args:
            target_count: 10_000
        - cellarium.ml.transforms.Log1p
      model: cellarium.ml.models.OnePassMeanVarStd
    data:
      dadc:
        class_path: cellarium.ml.data.DistributedAnnDataCollection
        init_args:
          filenames: https://storage.googleapis.com/dsp-cellarium-cas-public/test-data/test_{{0..1}}.h5ad
          shard_size: 100
          obs_columns_to_validate: [total_mrna_umis]
          max_cache_size: 2
      batch_keys:
        x_ng:
          attr: X
          convert_fn: cellarium.ml.utilities.data.densify
        var_names_g:
          attr: var_names
        total_mrna_umis_n:
          attr: obs
          key: total_mrna_umis
      batch_size: 50
      num_workers: 2
    trainer:
      accelerator: cpu
      devices: 1
      default_root_dir: {tmp_path}
    """
    with open(tmp_path / "onepass_config.yaml", "w") as f:
        f.write(onepass_config)
    main(["onepass_mean_var_std", "fit", "--config", str(tmp_path / "onepass_config.yaml")])
    ckpt_path = tmp_path / "lightning_logs" / "version_0" / "checkpoints" / "epoch=0-step=4.ckpt"

    lr_config = f"""
    model:
      transforms:
        - class_path: cellarium.ml.transforms.NormalizeTotal
          init_args:
            target_count: 10_000
        - cellarium.ml.transforms.Log1p
        - class_path: cellarium.ml.transforms.ZScore
          init_args:
            mean_g:
              !CheckpointLoader
              file_path: {ckpt_path}
              attr: model.mean_g
              convert_fn: null
            std_g:
              !CheckpointLoader
              file_path: {ckpt_path}
              attr: model.std_g
              convert_fn: null
            var_names_g:
              !CheckpointLoader
              file_path: {ckpt_path}
              attr: model.var_names_g
              convert_fn: numpy.ndarray.tolist
      model: cellarium.ml.models.LogisticRegression
      optim_fn: torch.optim.Adam
    data:
      dadc:
        class_path: cellarium.ml.data.DistributedAnnDataCollection
        init_args:
          filenames: https://storage.googleapis.com/dsp-cellarium-cas-public/test-data/test_{{0..1}}.h5ad
          shard_size: 100
          obs_columns_to_validate: [cell_type, total_mrna_umis]
          max_cache_size: 2
      batch_keys:
        x_ng:
          attr: X
          convert_fn: cellarium.ml.utilities.data.densify
        y_n:
          attr: obs
          key: cell_type
          convert_fn: cellarium.ml.utilities.data.categories_to_codes
        var_names_g:
          attr: var_names
        total_mrna_umis_n:
          attr: obs
          key: total_mrna_umis
      batch_size: 50
      num_workers: 2
    trainer:
      accelerator: cpu
      devices: 1
      max_steps: 1
    """
    with open(tmp_path / "lr_config.yaml", "w") as f:
        f.write(lr_config)
    main(["logistic_regression", "fit", "--config", str(tmp_path / "lr_config.yaml")])


def test_compute_var_names_g(tmp_path: Path) -> None:
    ipca_config = f"""
    model:
      transforms:
        - class_path: cellarium.ml.transforms.NormalizeTotal
          init_args:
            target_count: 10_000
        - cellarium.ml.transforms.Log1p
      model:
        class_path: cellarium.ml.models.IncrementalPCA
        init_args:
            n_components: 50
    data:
      dadc:
        class_path: cellarium.ml.data.DistributedAnnDataCollection
        init_args:
          filenames: https://storage.googleapis.com/dsp-cellarium-cas-public/test-data/test_{{0..1}}.h5ad
          shard_size: 100
          obs_columns_to_validate: [total_mrna_umis]
      batch_keys:
        x_ng:
          attr: X
          convert_fn: cellarium.ml.utilities.data.densify
        var_names_g:
          attr: var_names
        total_mrna_umis_n:
          attr: obs
          key: total_mrna_umis
      batch_size: 100
    trainer:
      accelerator: cpu
      devices: 1
      default_root_dir: {tmp_path}
    """
    with open(ipca_config_path := str(tmp_path / "ipca_config.yaml"), "w") as f:
        f.write(ipca_config)
    main(["incremental_pca", "fit", "--config", ipca_config_path])
    ckpt_path = tmp_path / "lightning_logs" / "version_0" / "checkpoints" / "epoch=0-step=2.ckpt"

    lr_config = f"""
    model:
      transforms:
        - !CheckpointLoader
          file_path: {ckpt_path}
          attr: null
          convert_fn: null
      model: cellarium.ml.models.LogisticRegression
      optim_fn: torch.optim.Adam
    data:
      dadc:
        class_path: cellarium.ml.data.DistributedAnnDataCollection
        init_args:
          filenames: https://storage.googleapis.com/dsp-cellarium-cas-public/test-data/test_{{0..1}}.h5ad
          shard_size: 100
          obs_columns_to_validate: [cell_type, total_mrna_umis]
      batch_keys:
        x_ng:
          attr: X
          convert_fn: cellarium.ml.utilities.data.densify
        y_n:
          attr: obs
          key: cell_type
          convert_fn: cellarium.ml.utilities.data.categories_to_codes
        var_names_g:
          attr: var_names
        total_mrna_umis_n:
          attr: obs
          key: total_mrna_umis
      batch_size: 100
    trainer:
      accelerator: cpu
      devices: 1
      max_steps: 1
    """
    with open(lr_config_path := str(tmp_path / "lr_config.yaml"), "w") as f:
        f.write(lr_config)
    main(["logistic_regression", "fit", "--config", lr_config_path])



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import math
import os
from pathlib import Path

import lightning.pytorch as pl
import numpy as np
import pytest
import torch

from cellarium.ml import CellariumModule
from cellarium.ml.models import Geneformer
from cellarium.ml.utilities.data import collate_fn
from tests.common import BoringDataset


def test_load_from_checkpoint_multi_device(tmp_path: Path):
    n, g = 4, 3
    var_names_g = [f"gene_{i}" for i in range(g)]
    devices = int(os.environ.get("TEST_DEVICES", "1"))
    # dataloader
    train_loader = torch.utils.data.DataLoader(
        BoringDataset(
            np.arange(n * g).reshape(n, g),
            var_names=np.array(var_names_g),
        ),
        collate_fn=collate_fn,
    )
    # model
    model = Geneformer(
        var_names_g=var_names_g,
        hidden_size=2,
        num_hidden_layers=1,
        num_attention_heads=1,
        intermediate_size=4,
        max_position_embeddings=2,
    )
    module = CellariumModule(model=model, optim_fn=torch.optim.Adam, optim_kwargs={"lr": 1e-3})
    # trainer
    trainer = pl.Trainer(
        accelerator="cpu",
        devices=devices,
        max_epochs=1,
        default_root_dir=tmp_path,
    )
    # fit
    trainer.fit(module, train_dataloaders=train_loader)

    # run tests only for rank 0
    if trainer.global_rank != 0:
        return

    # load model from checkpoint
    ckpt_path = tmp_path / f"lightning_logs/version_0/checkpoints/epoch=0-step={math.ceil(n / devices)}.ckpt"
    assert ckpt_path.is_file()
    loaded_model: Geneformer = CellariumModule.load_from_checkpoint(ckpt_path).model
    # assert
    assert np.array_equal(model.var_names_g, loaded_model.var_names_g)
    np.testing.assert_allclose(model.feature_ids, loaded_model.feature_ids)


@pytest.mark.parametrize("perturb", ["activation", "deletion", "map", "none"])
def test_tokenize_with_perturbations(perturb: str):
    var_names_g = ["a", "b", "c", "d"]
    geneformer = Geneformer(var_names_g=var_names_g)
    x_ng = torch.tensor([[4, 3, 2, 1]])  # sort order will be [a,b,c,d] and tokens will be [2,3,4,5]

    # test that we get the expected output for a well-formed set of input args
    match perturb:
        case "none":
            input_ids, _ = geneformer.tokenize_with_perturbations(x_ng)
            expected_input_ids = torch.tensor([[2, 3, 4, 5]])
        case "activation":
            input_ids, _ = geneformer.tokenize_with_perturbations(x_ng, feature_activation=["d"])
            expected_input_ids = torch.tensor([[5, 2, 3, 4]])
        case "deletion":
            input_ids, _ = geneformer.tokenize_with_perturbations(x_ng, feature_deletion=["c"])
            expected_input_ids = torch.tensor([[2, 3, 5, 0]])
        case "map":
            input_ids, _ = geneformer.tokenize_with_perturbations(x_ng, feature_map={"a": 1, "b": 1})
            expected_input_ids = torch.tensor([[1, 1, 4, 5]])

    print(f"Expected input_ids:\n{expected_input_ids}")
    print(f"Actual input_ids:\n{input_ids}")
    torch.testing.assert_close(input_ids, expected_input_ids)

    # test that we raise an AssertionError if we try to perturb something outside schema
    if perturb == "none":
        return
    with pytest.raises(ValueError):
        match perturb:
            case "activation":
                geneformer.tokenize_with_perturbations(x_ng, feature_activation=["e"])
            case "deletion":
                geneformer.tokenize_with_perturbations(x_ng, feature_deletion=["e"])
            case "map":
                geneformer.tokenize_with_perturbations(x_ng, feature_map={"a": 1, "e": 1})



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import math
from itertools import islice
from typing import Literal

import numpy as np
import torch
from anndata import AnnData
from boltons.iterutils import chunked_iter
from torch.utils.data import IterableDataset

from cellarium.ml.data.distributed_anndata import DistributedAnnDataCollection
from cellarium.ml.utilities.data import AnnDataField, get_rank_and_num_replicas, get_worker_info


class IterableDistributedAnnDataCollectionDataset(IterableDataset):
    r"""
    Iterable DistributedAnnDataCollection Dataset.

    When :attr:`shuffle` is set to ``True`` then the iterator yields datapoints that are
    uniformly sampled from the entire dataset. Typical use cases include training variational
    models using the stochastic gradient descent algorithm.

    In order to maximize buffer usage, we only shuffle shards and datapoints within individual
    shards (and not across shards). Therefore, to achieve unbiased pseudo-random uniform sampling,
    it is imperative that the shards themselves contain datapoints that are uniformly sampled
    from the entire dataset. If correlations exist between datapoints in a given shard (e.g. all
    cells coming from the same tissue or experiment), then this assumption is violated. It is
    the user's responsibility to prepare appropriately shuffled data shards.

    Example::

        >>> from cellarium.ml.data import (
        ...     DistributedAnnDataCollection,
        ...     IterableDistributedAnnDataCollectionDataset,
        ... )
        >>> from cellarium.ml.utilities.data import AnnDataField, densify

        >>> dadc = DistributedAnnDataCollection(
        ...     "gs://bucket-name/folder/adata{000..005}.h5ad",
        ...     shard_size=10_000,
        ...     max_cache_size=2)

        >>> dataset = IterableDistributedAnnDataCollectionDataset(
        ...     dadc,
        ...     batch_keys={
        ...         "x_ng": AnnDataField(attr="X", convert_fn=densify),
        ...         "var_names_g": AnnDataField(attr="var_names"),
        ...     },
        ...     batch_size=5000,
        ...     iteration_strategy="cache_efficient",
        ...     shuffle=True,
        ...     seed=0,
        ...     drop_last=True,
        ... )

    Args:
        dadc:
            DistributedAnnDataCollection or AnnData from which to load the data.
        batch_keys:
            Dictionary that specifies which attributes and keys of the :attr:`dadc` to return
            in the batch data and how to convert them. Keys must correspond to
            the input keys of the transforms or the model. Values must be instances of
            :class:`cellarium.ml.utilities.data.AnnDataField`.
        batch_size:
            How many samples per batch to load.
        iteration_strategy:
            Strategy to use for iterating through the dataset. Options are ``same_order`` and ``cache_efficient``.
            ``same_order`` will iterate through the dataset in the same order independent of the number of replicas
            and workers. ``cache_efficient`` will try to minimize the amount of anndata files fetched by each worker.
        shuffle:
            If ``True``, the data is reshuffled at every epoch.
        seed:
            Random seed used to shuffle the sampler if :attr:`shuffle=True`.
        drop_last:
            If ``True``, then the sampler will drop the tail of the data
            to make it evenly divisible across the number of replicas. If ``False``,
            the sampler will add extra indices to make the data evenly divisible across
            the replicas.
        start_idx:
            The starting index of the dataset. If ``None``, then the dataset will start from the first index.
        end_idx:
            The ending index (exclusive) of the dataset. If ``None``, then the dataset will end at
            the last index (inclusive).
        test_mode:
            If ``True``, then tracking of cache and worker informations will be enabled.
    """

    def __init__(
        self,
        dadc: DistributedAnnDataCollection | AnnData,
        batch_keys: dict[str, AnnDataField],
        batch_size: int = 1,
        iteration_strategy: Literal["same_order", "cache_efficient"] = "cache_efficient",
        shuffle: bool = False,
        seed: int = 0,
        drop_last: bool = False,
        start_idx: int | None = None,
        end_idx: int | None = None,
        test_mode: bool = False,
    ) -> None:
        self.dadc = dadc
        if isinstance(dadc, AnnData):
            # mimic a DistributedAnnDataCollection
            self.dadc.limits = [dadc.n_obs]
        self.batch_keys = batch_keys
        self.batch_size = batch_size
        self.iteration_strategy = iteration_strategy
        self.shuffle = shuffle
        self.seed = seed
        self.drop_last = drop_last
        self.start_idx = 0 if start_idx is None else start_idx
        self.end_idx = dadc.n_obs if end_idx is None else end_idx
        self.epoch = 0
        self.test_mode = test_mode

    def __len__(self) -> int:
        """
        Returns the number of batches per replica.
        """
        _, num_replicas = get_rank_and_num_replicas()

        n_obs = self.end_idx - self.start_idx
        if self.drop_last and n_obs % num_replicas != 0:
            # Split to nearest available length that is evenly divisible.
            # This is to ensure each rank receives the same amount of data.
            per_replica = n_obs // num_replicas
        else:
            per_replica = math.ceil(n_obs / num_replicas)
        return math.ceil(per_replica / float(self.batch_size))

    def set_epoch(self, epoch: int) -> None:
        r"""
        Sets the epoch for the iterator. When :attr:`shuffle=True`, this ensures all replicas
        use a different random ordering for each epoch.
        """
        self.epoch = epoch

    def __getitem__(self, idx: int | list[int] | slice) -> dict[str, np.ndarray]:
        r"""
        Returns a dictionary containing the data from the :attr:`dadc` with keys specified by the :attr:`batch_keys`
        at the given index ``idx``.
        """

        data = {}
        for key, field in self.batch_keys.items():
            data[key] = field(self.dadc, idx)

        # for testing purposes
        if self.test_mode:
            rank, num_replicas = get_rank_and_num_replicas()
            worker_id, num_workers = get_worker_info()
            data["rank"] = np.array([rank])
            data["num_replicas"] = np.array([num_replicas])
            data["worker_id"] = np.array([worker_id])
            data["num_workers"] = np.array([num_workers])
            data["miss_count"] = np.array([self.dadc.cache.miss_count])
            data["epoch"] = np.array([self.epoch])

        return data

    def __iter__(self):
        r"""
        Iterate through the dataset by trying to minimize the amount of anndata files fetched by each worker.
        Iterated indices are evenly divided between replicas (see :attr:`drop_last`).

        .. note::

            1. For both strategies the amount of anndata files fetched is reduced by
               shuffling the shards first and then the datapoints within the shards.
            2. ``same_order`` strategy will iterate through the dataset in the same order independent
               of the number of replicas and workers.
            3. For ``cache_efficient`` strategy the amount of anndata files fetched is further
               reduced by assigning to each worker a contiguous chunk of the dataset.
               The returned iterator is determined by the ``torch.utils.data.get_worker_info()``
               and ``torch.distributed`` contexts.

        **Example 1**::

            indices=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
            num_replicas=1
            batch_size=2
            num_workers=3

        Same order:

        +------------+-------+-------+-------+-------+-------+---------+
        | batch idx  | 0     | 1     | 2     | 3     | 4     | 5       |
        +============+=======+=======+=======+=======+=======+=========+
        | indices    | (0,1) | (2,3) | (4,5) | (6,7) | (8,9) | (10,11) |
        +------------+-------+-------+-------+-------+-------+---------+
        | worker id  | 0     | 1     | 2     | 0     | 1     | 2       |
        +------------+-------+-------+-------+-------+-------+---------+

        Cache efficient:

        +------------+-------+-------+-------+-------+-------+---------+
        | batch idx  | 0     | 1     | 2     | 3     | 4     | 5       |
        +============+=======+=======+=======+=======+=======+=========+
        | indices    | (0,1) | (4,5) | (8,9) | (2,3) | (6,7) | (10,11) |
        +------------+-------+-------+-------+-------+-------+---------+
        | worker id  | 0     | 1     | 2     | 0     | 1     | 2       |
        +------------+-------+-------+-------+-------+-------+---------+


        **Example 2**::

            indices=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
            num_replicas=1
            batch_size=2
            num_workers=2

        Same order:

        +------------+-------+-------+-------+-------+-------+---------+
        | batch idx  | 0     | 1     | 2     | 3     | 4     | 5       |
        +============+=======+=======+=======+=======+=======+=========+
        | indices    | (0,1) | (2,3) | (4,5) | (6,7) | (8,9) | (10,)   |
        +------------+-------+-------+-------+-------+-------+---------+
        | worker id  | 0     | 1     | 0     | 1     | 0     | 1       |
        +------------+-------+-------+-------+-------+-------+---------+

        Cache efficient:

        +------------+-------+-------+-------+-------+-------+---------+
        | batch idx  | 0     | 1     | 2     | 3     | 4     | 5       |
        +============+=======+=======+=======+=======+=======+=========+
        | indices    | (0,1) | (6,7) | (2,3) | (8,9) | (4,5) | (10,)   |
        +------------+-------+-------+-------+-------+-------+---------+
        | worker id  | 0     | 1     | 0     | 1     | 0     | 1       |
        +------------+-------+-------+-------+-------+-------+---------+

        **Example 3**::

            indices=[0, 1, 2, 3, 4, 5, 6, 7]
            num_replicas=1
            batch_size=3
            num_workers=2

        Same order:

        +------------+---------+---------+-------+
        | batch idx  | 0       | 1       | 2     |
        +============+=========+=========+=======+
        | indices    | (0,1,2) | (3,4,5) | (6,7) |
        +------------+---------+---------+-------+
        | worker id  | 0       | 1       | 0     |
        +------------+---------+---------+-------+

        Cache efficient:

        +------------+---------+-------+---------+
        | batch idx  | 0       | 1     | 2       |
        +============+=========+=======+=========+
        | indices    | (0,1,2) | (6,7) | (3,4,5) |
        +------------+---------+-------+---------+
        | worker id  | 0       | 1     | 0       |
        +------------+---------+-------+---------+

        **Example 4**::

            indices=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
            num_replicas=2
            drop_last=True
            batch_size=2
            num_workers=1

        Same order:

        *Replica 1*

        +------------+-------+-------+------+
        | batch idx  | 0     | 1     | 2    |
        +============+=======+=======+======+
        | indices    | (0,2) | (4,6) | (8,) |
        +------------+-------+-------+------+
        | worker id  | 0     | 0     | 0    |
        +------------+-------+-------+------+

        *Replica 2*

        +------------+-------+-------+------+
        | batch idx  | 0     | 1     | 2    |
        +============+=======+=======+======+
        | indices    | (1,3) | (5,7) | (9,) |
        +------------+-------+-------+------+
        | worker id  | 0     | 0     | 0    |
        +------------+-------+-------+------+

        Cache efficient:

        *Replica 1*

        +------------+-------+-------+------+
        | batch idx  | 0     | 1     | 2    |
        +============+=======+=======+======+
        | indices    | (0,1) | (2,3) | (4,) |
        +------------+-------+-------+------+
        | worker id  | 0     | 0     | 0    |
        +------------+-------+-------+------+

        *Replica 2*

        +------------+-------+-------+------+
        | batch idx  | 0     | 1     | 2    |
        +============+=======+=======+======+
        | indices    | (5,6) | (7,8) | (9,) |
        +------------+-------+-------+------+
        | worker id  | 0     | 0     | 0    |
        +------------+-------+-------+------+


        **Example 5**::

            indices=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
            num_replicas=2
            drop_last=False
            batch_size=2
            num_workers=1

        Same order:

        *Replica 1*

        +------------+-------+-------+--------+
        | batch idx  | 0     | 1     | 2      |
        +============+=======+=======+========+
        | indices    | (0,2) | (4,6) | (8,10) |
        +------------+-------+-------+--------+
        | worker id  | 0     | 0     | 0      |
        +------------+-------+-------+--------+

        *Replica 2*

        +------------+-------+-------+-------+
        | batch idx  | 0     | 1     | 2     |
        +============+=======+=======+=======+
        | indices    | (1,3) | (5,7) | (9,0) |
        +------------+-------+-------+-------+
        | worker id  | 0     | 0     | 0     |
        +------------+-------+-------+-------+

        Cache efficient:

        *Replica 1*

        +------------+-------+-------+-------+
        | batch idx  | 0     | 1     | 2     |
        +============+=======+=======+=======+
        | indices    | (0,1) | (2,3) | (4,5) |
        +------------+-------+-------+-------+
        | worker id  | 0     | 0     | 0     |
        +------------+-------+-------+-------+

        *Replica 2*

        +------------+-------+-------+--------+
        | batch idx  | 0     | 1     | 2      |
        +============+=======+=======+========+
        | indices    | (6,7) | (8,9) | (10,0) |
        +------------+-------+-------+--------+
        | worker id  | 0     | 0     | 0      |
        +------------+-------+-------+--------+
        """
        if self.test_mode and isinstance(self.dadc, DistributedAnnDataCollection):
            # clear lru cache
            self.dadc.cache.clear()

        # replicas
        rank, num_replicas = get_rank_and_num_replicas()

        if self.drop_last and len(self.dadc) % num_replicas != 0:
            # Split to nearest available length that is evenly divisible.
            # This is to ensure each rank receives the same amount of data.
            per_replica = len(self.dadc) // num_replicas
        else:
            per_replica = math.ceil(len(self.dadc) / num_replicas)
        total_size = per_replica * num_replicas

        # workers
        worker_id, num_workers = get_worker_info()

        # indices
        if self.shuffle:
            rng = torch.Generator()
            rng.manual_seed(self.seed + self.epoch)
            limits = [idx for idx in self.dadc.limits if idx > self.start_idx and idx < self.end_idx]
            iter_limits = list(zip([self.start_idx] + limits, limits + [self.end_idx]))
            # shuffle shards
            limit_indices = torch.randperm(len(iter_limits), generator=rng).tolist()
            indices = []
            for limit_idx in limit_indices:
                lower, upper = iter_limits[limit_idx]
                # shuffle cells within shards
                indices.extend((torch.randperm(upper - lower, generator=rng) + lower).tolist())
        else:
            indices = list(range(self.start_idx, self.end_idx))

        if not self.drop_last:
            # add extra samples to make it evenly divisible
            padding_size = total_size - len(indices)
            if padding_size <= len(indices):
                indices += indices[:padding_size]
            else:
                indices += (indices * math.ceil(padding_size / len(indices)))[:padding_size]
        else:
            # remove tail of data to make it evenly divisible.
            indices = indices[:total_size]

        if self.iteration_strategy == "same_order":
            # replica indices
            indices = indices[rank:total_size:num_replicas]
            if len(indices) != per_replica:
                raise ValueError(
                    f"The number of indices must be equal to the per_replica size. "
                    f"Got {len(indices)} != {per_replica} at rank {rank}."
                )

            # in python 3.12 `chunked_iter` can be replaced with `itertools.batched`
            for batch_indices in islice(chunked_iter(indices, self.batch_size), worker_id, None, num_workers):
                yield self[batch_indices]

        elif self.iteration_strategy == "cache_efficient":
            # replica indices
            indices = indices[rank * per_replica : (rank + 1) * per_replica]
            if len(indices) != per_replica:
                raise ValueError(
                    f"The number of indices must be equal to the per_replica size. "
                    f"Got {len(indices)} != {per_replica} at rank {rank}."
                )

            # worker indices
            batches_per_replica = math.ceil(per_replica / float(self.batch_size))
            batches_per_worker = math.ceil(batches_per_replica / float(num_workers))
            per_worker = batches_per_worker * self.batch_size

            iter_start = worker_id * per_worker
            iter_end = min(iter_start + per_worker, per_replica)
            indices = indices[iter_start:iter_end]

            # in python 3.12 `chunked_iter` can be replaced with `itertools.batched`
            for batch_indices in chunked_iter(indices, self.batch_size):
                yield self[batch_indices]

        # Sets epoch for persistent workers
        self.set_epoch(self.epoch + 1)



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

"""
Command line interface for Cellarium ML.
"""

import copy
import sys
import warnings
from collections.abc import Callable
from dataclasses import dataclass
from functools import cache
from operator import attrgetter
from typing import Any

import numpy as np
import torch
import yaml
from jsonargparse import Namespace, class_from_function
from jsonargparse._loaders_dumpers import DefaultLoader
from jsonargparse._util import import_object
from lightning.pytorch.cli import ArgsType, LightningArgumentParser, LightningCLI
from torch._subclasses.fake_tensor import FakeCopyMode, FakeTensorMode

from cellarium.ml import CellariumAnnDataDataModule, CellariumModule, CellariumPipeline
from cellarium.ml.utilities.data import collate_fn

cached_loaders = {}


@dataclass
class FileLoader:
    """
    A YAML constructor for loading a file and accessing its attributes.

    Example:

    .. code-block:: yaml

        model:
          transforms:
            - class_path: cellarium.ml.transforms.Filter
              init_args:
                filter_list:
                  !FileLoader
                  file_path: gs://dsp-cellarium-cas-public/test-data/filter_list.csv
                  loader_fn: pandas.read_csv
                  attr: index
                  convert_fn: numpy.ndarray.tolist

    Args:
        file_path:
            The file path to load the object from.
        loader_fn:
            A function to load the object from the file path.
        attr:
            An attribute to get from the loaded object. If ``None`` the loaded object is returned.
        convert_fn:
            A function to convert the loaded object. If ``None`` the loaded object is returned.
    """

    file_path: str
    loader_fn: Callable[[str], Any] | str
    attr: str | None = None
    convert_fn: Callable[[Any], Any] | str | None = None

    def __new__(cls, file_path, loader_fn, attr, convert_fn):
        if isinstance(loader_fn, str):
            loader_fn = import_object(loader_fn)
        if loader_fn not in cached_loaders:
            cached_loaders[loader_fn] = cache(loader_fn)
        loader_fn = cached_loaders[loader_fn]
        obj = loader_fn(file_path)

        if attr is not None:
            obj = attrgetter(attr)(obj)

        if isinstance(convert_fn, str):
            convert_fn = import_object(convert_fn)
        if convert_fn is not None:
            obj = convert_fn(obj)

        return obj


@dataclass
class CheckpointLoader(FileLoader):
    """
    A YAML constructor for loading a :class:`~cellarium.ml.core.CellariumModule` checkpoint and accessing its
    attributes.

    Example:

    .. code-block:: yaml

        model:
          transorms:
            - class_path: cellarium.ml.transforms.DivideByScale
              init_args:
                scale_g:
                  !CheckpointLoader
                  file_path: gs://dsp-cellarium-cas-public/test-data/tdigest.ckpt
                  attr: model.median_g
                  convert_fn: null

    Args:
        file_path:
            The file path to load the object from.
        attr:
            An attribute to get from the loaded object. If ``None`` the loaded object is returned.
        convert_fn:
            A function to convert the loaded object. If ``None`` the loaded object is returned.
    """

    file_path: str
    attr: str | None = None
    convert_fn: Callable[[Any], Any] | str | None = None

    def __new__(cls, file_path, attr, convert_fn):
        return super().__new__(cls, file_path, CellariumModule.load_from_checkpoint, attr, convert_fn)


def file_loader_constructor(loader: yaml.SafeLoader, node: yaml.nodes.MappingNode) -> FileLoader:
    """Construct an object from a file."""
    return FileLoader(**loader.construct_mapping(node))  # type: ignore[misc]


def checkpoint_loader_constructor(loader: yaml.SafeLoader, node: yaml.nodes.MappingNode) -> CheckpointLoader:
    """Construct an object from a checkpoint."""
    return CheckpointLoader(**loader.construct_mapping(node))  # type: ignore[misc]


loader = DefaultLoader
loader.add_constructor("!FileLoader", file_loader_constructor)
loader.add_constructor("!CheckpointLoader", checkpoint_loader_constructor)


REGISTERED_MODELS = {}


def register_model(model: Callable[[ArgsType], None]):
    REGISTERED_MODELS[model.__name__] = model
    return model


CellariumModuleLoadFromCheckpoint = class_from_function(CellariumModule.load_from_checkpoint, CellariumModule)


@dataclass
class LinkArguments:
    """
    Arguments for linking the value of a target argument to the values of one or more source arguments.

    Args:
        source:
            Key(s) from which the target value is derived.
        target:
            Key to where the value is set.
        compute_fn:
            Function to compute target value from source.
        apply_on:
            At what point to set target value, ``"parse"`` or ``"instantiate"``.
    """

    source: str | tuple[str, ...]
    target: str
    compute_fn: Callable | None = None
    apply_on: str = "instantiate"


def compute_n_obs(data: CellariumAnnDataDataModule) -> int:
    """
    Compute the number of observations in the data.

    Args:
        data: A :class:`CellariumAnnDataDataModule` instance.

    Returns:
        The number of observations in the data.
    """
    return data.dadc.n_obs


def compute_n_categories(data: CellariumAnnDataDataModule) -> int:
    """
    Compute the number of categories in the target variable.

    E.g. if the target variable is ``obs["cell_type"]`` then this function
    returns the number of categories in ``obs["cell_type"]``::

        >>> len(data.dadc[0].obs["cell_type"].cat.categories)

    Args:
        data: A :class:`CellariumAnnDataDataModule` instance.

    Returns:
        The number of categories in the target variable.
    """
    field = data.batch_keys["y_n"]
    value = getattr(data.dadc[0], field.attr)
    if field.key is not None:
        value = value[field.key]
    return len(value.cat.categories)


def compute_var_names_g(transforms: list[torch.nn.Module], data: CellariumAnnDataDataModule) -> np.ndarray:
    """
    Compute variable names from the data by applying the transforms.

    Args:
        transforms:
            A list of transforms.
        data:
            A :class:`CellariumAnnDataDataModule` instance.

    Returns:
        The variable names.
    """
    batch = {key: field(data.dadc, 0) for key, field in data.batch_keys.items()}
    pipeline = CellariumPipeline(transforms)
    with FakeTensorMode(allow_non_fake_inputs=True) as fake_mode:
        fake_batch = collate_fn([batch])
        with FakeCopyMode(fake_mode):
            fake_pipeline = copy.deepcopy(pipeline)
        fake_pipeline.to("cpu")
        output = fake_pipeline(fake_batch)
    return output["var_names_g"]


def lightning_cli_factory(
    model_class_path: str,
    link_arguments: list[LinkArguments] | None = None,
    trainer_defaults: dict[str, Any] | None = None,
) -> type[LightningCLI]:
    """
    Factory function for creating a :class:`LightningCLI` with a preset model and custom argument linking.

    Example::

        cli = lightning_cli_factory(
            "cellarium.ml.models.IncrementalPCA",
            link_arguments=[
                LinkArguments(("model.transforms", "data"), "model.model.init_args.var_names_g", compute_var_names_g)
            ],
            trainer_defaults={
                "max_epochs": 1,  # one pass
                "strategy": {
                    "class_path": "lightning.pytorch.strategies.DDPStrategy",
                    "dict_kwargs": {"broadcast_buffers": False},
                },
            },
        )

    Args:
        model_class_path:
            A string representation of the model class path (e.g., ``"cellarium.ml.models.IncrementalPCA"``).
        link_arguments:
            A list of :class:`LinkArguments` that specify how to derive the value of a target
            argument from the values of one or more source arguments. If ``None`` then no
            arguments are linked.
        trainer_defaults:
            Default values for the trainer.

    Returns:
        A :class:`LightningCLI` class with the given model and argument linking.
    """

    class NewLightningCLI(LightningCLI):
        def __init__(self, args: ArgsType = None) -> None:
            super().__init__(
                CellariumModule,
                CellariumAnnDataDataModule,
                trainer_defaults=trainer_defaults,
                args=args,
            )

        def instantiate_classes(self) -> None:
            with torch.device("meta"):
                # skip the initialization of model parameters
                # parameters are later initialized by the  `CellariumModule.configure_model` method
                return super().instantiate_classes()

        def add_arguments_to_parser(self, parser: LightningArgumentParser) -> None:
            if link_arguments is not None:
                for link in link_arguments:
                    parser.link_arguments(link.source, link.target, link.compute_fn, link.apply_on)
            # this is helpful for generating a default config file with --print_config
            parser.set_defaults(
                {
                    "model.model": model_class_path,
                    "data.dadc": "cellarium.ml.data.DistributedAnnDataCollection",
                }
            )

    return NewLightningCLI


@register_model
def geneformer(args: ArgsType = None) -> None:
    r"""
    CLI to run the :class:`cellarium.ml.models.Geneformer` model.

    This example shows how to fit feature count data to the Geneformer model [1].

    Example run::

        cellarium-ml geneformer fit \
            --data.filenames "gs://dsp-cellarium-cas-public/test-data/test_{0..3}.h5ad" \
            --data.shard_size 100 \
            --data.max_cache_size 2 \
            --data.batch_size 5 \
            --data.num_workers 1 \
            --trainer.accelerator gpu \
            --trainer.devices 1 \
            --trainer.default_root_dir runs/geneformer \
            --trainer.max_steps 10

    **References:**

    1. `Transfer learning enables predictions in network biology (Theodoris et al.)
       <https://www.nature.com/articles/s41586-023-06139-9>`_.

    Args:
        args: Arguments to parse. If ``None`` the arguments are taken from ``sys.argv``.
    """
    cli = lightning_cli_factory(
        "cellarium.ml.models.Geneformer",
        link_arguments=[
            LinkArguments(("model.transforms", "data"), "model.model.init_args.var_names_g", compute_var_names_g)
        ],
    )
    cli(args=args)


@register_model
def incremental_pca(args: ArgsType = None) -> None:
    r"""
    CLI to run the :class:`cellarium.ml.models.IncrementalPCA` model.

    This example shows how to fit feature count data to incremental PCA
    model [1, 2].

    Example run::

        cellarium-ml incremental_pca fit \
            --model.model.init_args.n_components 50 \
            --data.filenames "gs://dsp-cellarium-cas-public/test-data/test_{0..3}.h5ad" \
            --data.shard_size 100 \
            --data.max_cache_size 2 \
            --data.batch_size 100 \
            --data.num_workers 4 \
            --trainer.accelerator gpu \
            --trainer.devices 1 \
            --trainer.default_root_dir runs/ipca \

    **References:**

    1. `A Distributed and Incremental SVD Algorithm for Agglomerative Data Analysis on Large Networks (Iwen et al.)
       <https://users.math.msu.edu/users/iwenmark/Papers/distrib_inc_svd.pdf>`_.
    2. `Incremental Learning for Robust Visual Tracking (Ross et al.)
       <https://www.cs.toronto.edu/~dross/ivt/RossLimLinYang_ijcv.pdf>`_.

    Args:
        args: Arguments to parse. If ``None`` the arguments are taken from ``sys.argv``.
    """
    cli = lightning_cli_factory(
        "cellarium.ml.models.IncrementalPCA",
        link_arguments=[
            LinkArguments(("model.transforms", "data"), "model.model.init_args.var_names_g", compute_var_names_g)
        ],
        trainer_defaults={
            "max_epochs": 1,  # one pass
            "strategy": {
                "class_path": "lightning.pytorch.strategies.DDPStrategy",
                "dict_kwargs": {"broadcast_buffers": False},
            },
        },
    )
    cli(args=args)


@register_model
def logistic_regression(args: ArgsType = None) -> None:
    r"""
    CLI to run the :class:`cellarium.ml.models.LogisticRegression` model.

    Example run::

        cellarium-ml logistic_regression fit \
            --data.filenames "gs://dsp-cellarium-cas-public/test-data/test_{0..3}.h5ad" \
            --data.shard_size 100 \
            --data.max_cache_size 2 \
            --data.batch_keys.x_ng.attr X \
            --data.batch_keys.x_ng.convert_fn cellarium.ml.utilities.data.densify \
            --data.batch_keys.var_names_g.attr var_names \
            --data.batch_keys.y_n.attr obs \
            --data.batch_keys.y_n.key cell_type \
            --data.batch_keys.y_n.convert_fn cellarium.ml.utilities.data.categories_to_codes \
            --data.batch_size 100 \
            --data.num_workers 4 \
            --trainer.accelerator gpu \
            --trainer.devices 1 \
            --trainer.max_steps 1000

    Args:
        args: Arguments to parse. If ``None`` the arguments are taken from ``sys.argv``.
    """

    cli = lightning_cli_factory(
        "cellarium.ml.models.LogisticRegression",
        link_arguments=[
            LinkArguments(("model.transforms", "data"), "model.model.init_args.var_names_g", compute_var_names_g),
            LinkArguments("data", "model.model.init_args.n_obs", compute_n_obs),
            LinkArguments("data", "model.model.init_args.n_categories", compute_n_categories),
        ],
    )
    cli(args=args)


@register_model
def onepass_mean_var_std(args: ArgsType = None) -> None:
    r"""
    CLI to run the :class:`cellarium.ml.models.OnePassMeanVarStd` model.

    This example shows how to calculate mean, variance, and standard deviation of log normalized
    feature count data in one pass [1].

    Example run::

        cellarium-ml onepass_mean_var_std fit \
            --data.filenames "gs://dsp-cellarium-cas-public/test-data/test_{0..3}.h5ad" \
            --data.shard_size 100 \
            --data.max_cache_size 2 \
            --data.batch_size 100 \
            --data.num_workers 4 \
            --trainer.accelerator gpu \
            --trainer.devices 1 \
            --trainer.default_root_dir runs/onepass \

    **References:**

    1. `Algorithms for calculating variance
       <https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance>`_.

    Args:
        args: Arguments to parse. If ``None`` the arguments are taken from ``sys.argv``.
    """
    cli = lightning_cli_factory(
        "cellarium.ml.models.OnePassMeanVarStd",
        link_arguments=[
            LinkArguments(("model.transforms", "data"), "model.model.init_args.var_names_g", compute_var_names_g)
        ],
        trainer_defaults={
            "max_epochs": 1,  # one pass
            "strategy": {
                "class_path": "lightning.pytorch.strategies.DDPStrategy",
                "dict_kwargs": {"broadcast_buffers": False},
            },
        },
    )
    cli(args=args)


@register_model
def probabilistic_pca(args: ArgsType = None) -> None:
    r"""
    CLI to run the :class:`cellarium.ml.models.ProbabilisticPCA` model.

    This example shows how to fit feature count data to probabilistic PCA
    model [1].

    There are two flavors of probabilistic PCA model that are available:

    1. ``marginalized`` - latent variable ``z`` is marginalized out [1]. Marginalized
       model provides a closed-form solution for the marginal log-likelihood.
       Closed-form solution for the marginal log-likelihood has reduced
       variance compared to the ``linear_vae`` model.
    2. ``linear_vae`` - latent variable ``z`` has a diagonal Gaussian distribution [2].
       Training a linear VAE with variational inference recovers a uniquely identifiable
       global maximum  corresponding to the principal component directions.
       The global maximum of the ELBO objective for the linear VAE  is identical
       to the global maximum for the marginal log-likelihood of probabilistic PCA.

    Example run::

        cellarium-ml probabilistic_pca fit \
            --model.model.init_args.n_components 256 \
            --model.model.init_args.ppca_flavor marginalized \
            --data.filenames "gs://dsp-cellarium-cas-public/test-data/test_{0..3}.h5ad" \
            --data.shard_size 100 \
            --data.max_cache_size 2 \
            --data.batch_size 100 \
            --data.shuffle true \
            --data.num_workers 4 \
            --trainer.accelerator gpu \
            --trainer.devices 1 \
            --trainer.max_steps 1000 \
            --trainer.default_root_dir runs/ppca \

    **References:**

    1. `Probabilistic Principal Component Analysis (Tipping et al.)
       <https://www.robots.ox.ac.uk/~cvrg/hilary2006/ppca.pdf>`_.
    2. `Understanding Posterior Collapse in Generative Latent Variable Models (Lucas et al.)
       <https://openreview.net/pdf?id=r1xaVLUYuE>`_.

    Args:
        args: Arguments to parse. If ``None`` the arguments are taken from ``sys.argv``.
    """
    cli = lightning_cli_factory(
        "cellarium.ml.models.ProbabilisticPCA",
        link_arguments=[
            LinkArguments(("model.transforms", "data"), "model.model.init_args.var_names_g", compute_var_names_g),
            LinkArguments("data", "model.model.init_args.n_obs", compute_n_obs),
        ],
    )
    cli(args=args)


@register_model
def tdigest(args: ArgsType = None) -> None:
    r"""
    CLI to run the :class:`cellarium.ml.models.TDigest` model.

    This example shows how to calculate non-zero median of normalized feature count
    data in one pass [1].

    Example run::

        cellarium-ml tdigest fit \
            --data.filenames "gs://dsp-cellarium-cas-public/test-data/test_{0..3}.h5ad" \
            --data.shard_size 100 \
            --data.max_cache_size 2 \
            --data.batch_size 100 \
            --data.num_workers 4 \
            --trainer.accelerator cpu \
            --trainer.devices 1 \
            --trainer.default_root_dir runs/tdigest \

    **References:**

    1. `Computing Extremely Accurate Quantiles Using T-Digests (Dunning et al.)
       <https://github.com/tdunning/t-digest/blob/master/docs/t-digest-paper/histo.pdf>`_.

    Args:
        args: Arguments to parse. If ``None`` the arguments are taken from ``sys.argv``.
    """
    cli = lightning_cli_factory(
        "cellarium.ml.models.TDigest",
        link_arguments=[
            LinkArguments(("model.transforms", "data"), "model.model.init_args.var_names_g", compute_var_names_g)
        ],
        trainer_defaults={
            "max_epochs": 1,  # one pass
        },
    )
    cli(args=args)


def main(args: ArgsType = None) -> None:
    """
    CLI that dispatches to the appropriate model cli based on the model name in ``args`` and runs it.

    Args:
        args: Arguments to parse. If ``None`` the arguments are taken from ``sys.argv``.
            The model name is expected to be the first argument if ``args`` is a list
            or the ``model_name`` key if ``args`` is a dictionary or ``Namespace``.
    """
    if isinstance(args, (dict, Namespace)):
        if "model_name" not in args:
            raise ValueError("'model_name' key must be specified in args")
        model_name = args.pop("model_name")
    elif isinstance(args, list):
        if len(args) == 0:
            raise ValueError("'model_name' must be specified as the first argument in args")
        model_name = args.pop(0)
    elif args is None:
        args = sys.argv[1:].copy()
        if len(args) == 0:
            raise ValueError("'model_name' must be specified after cellarium-ml")
        model_name = args.pop(0)

    if model_name not in REGISTERED_MODELS:
        raise ValueError(f"'model_name' must be one of {list(REGISTERED_MODELS.keys())}. Got '{model_name}'")
    model_cli = REGISTERED_MODELS[model_name]
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="Transforming to str index.")
        warnings.filterwarnings("ignore", message="LightningCLI's args parameter is intended to run from within Python")
        warnings.filterwarnings("ignore", message="Your `IterableDataset` has `__len__` defined.")
        model_cli(args)  # run the model


if __name__ == "__main__":
    main()



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Callable
from dataclasses import dataclass
from typing import Literal

import torch
import torch.nn as nn


@dataclass
class abcdParameter:
    """
    An *abcd*-parametrization describes the scaling of a parameter :math:`W` with width :math:`n`. The parameter
    is initialized with a standard deviation :math:`\\sigma` and a parameter-specific learning rate scaling factor
    :math:`\\alpha` at base width :math:`n_0`. The scaling of the parameterization with width :math:`n` is described
    by a set of numbers :math:`\\{a, b, c, d\\}` such that:

    a. Parameter is given as :math:`W = \\sqrt{\\alpha} \\cdot (n_0 / n)^a \\cdot w` where :math:`w`
       is the learnable parameter.
    b. Learnable parameter is initialized as
       :math:`w \\sim \\mathcal{N}(0, \\sigma \\cdot (n_0 / n)^b / \\sqrt{\\alpha})`.
    c. The effective learning rate for :math:`W` is
       :math:`\\alpha \\cdot (n_0 / n)^{2a} \\cdot (n_0 / n)^c \\cdot \\eta` for some global learning rate
       :math:`\\eta`. In this implementation, :math:`c` is equal to :math:`0`.
    d. The gradients of :math:`w` are scaled by :math:`(n / n_0)^d`.

    Args:
        data:
            The tensor data.
        width:
            The width :math:`n` of the tensor.
        a:
            The :math:`a` parameter.
        b:
            The :math:`b` parameter.
        d:
            The :math:`d` parameter.
        init_std:
            The initialization standard deviation :math:`\\sigma` at base width :math:`n_0`.
        lr_scale:
            The learning rate scale factor :math:`\\alpha` at base width :math:`n_0`.
        base_width:
            The base width :math:`n_0`.
    """

    data: torch.Tensor | None = None
    width: int = 1
    a: float = 0.0
    b: float = 0.0
    d: float = 0.0
    init_std: float = 1.0
    lr_scale: float = 1.0
    base_width: int = 1


class MuLinear(nn.Module):
    r"""
    Linear layer with a maximal update parametrization.

    The maximal update parametrization for SGD is defined by:

    +-----------+----------------+----------------+----------------+
    |           | Input & Biases | Hidden         | Output         |
    +-----------+----------------+----------------+----------------+
    | :math:`a` | :math:`-0.5`   | :math:`0`      | :math:`0.5`    |
    +-----------+----------------+----------------+----------------+
    | :math:`b` | :math:`0.5`    | :math:`0.5`    | :math:`0.5`    |
    +-----------+----------------+----------------+----------------+
    | :math:`c` | :math:`0`      | :math:`0`      | :math:`0`      |
    +-----------+----------------+----------------+----------------+
    | :math:`d` | :math:`0`      | :math:`0`      | :math:`0`      |
    +-----------+----------------+----------------+----------------+
    | :math:`n` | out_features   | in_features    | in_features    |
    +-----------+----------------+----------------+----------------+

    The maximal update parametrization for Adam and AdamW is defined by

    +-----------+----------------+----------------+----------------+
    |           | Input & Biases | Hidden         | Output         |
    +-----------+----------------+----------------+----------------+
    | :math:`a` | :math:`0`      | :math:`1`      | :math:`1`      |
    +-----------+----------------+----------------+----------------+
    | :math:`b` | :math:`0`      | :math:`-0.5`   | :math:`0`      |
    +-----------+----------------+----------------+----------------+
    | :math:`c` | :math:`0`      | :math:`0`      | :math:`0`      |
    +-----------+----------------+----------------+----------------+
    | :math:`d` | :math:`1`      | :math:`2`      | :math:`1`      |
    +-----------+----------------+----------------+----------------+
    | :math:`n` | out_features   | in_features    | in_features    |
    +-----------+----------------+----------------+----------------+

    Since in this implementation :math:`c` always equals 0, regular PyTorch optimizers
    can be used.

    **References:**

    1. `Feature Learning in Infinite-Width Neural Networks (Yang et al.)
       <https://arxiv.org/pdf/2011.14522.pdf>`_.
    2. `Tensor Programs IVb: Adaptive Optimization in the ∞-Width Limit (Yang et al.)
       <https://arxiv.org/pdf/2308.01814.pdf>`_.

    Args:
        in_features:
            Size of each input sample.
        out_features:
            Size of each output sample.
        bias:
            If set to ``False``, the layer will not learn an additive bias.
        layer:
            Layer type.
        optimizer:
            Optimizer type.
        weight_init_std:
            The standard deviation of the weight initialization at base width.
        bias_init_std:
            The standard deviation of the bias initialization at base width.
        lr_scale:
            The learning rate scaling factor for the weight and the bias.
        base_width:
            The base width of the layer.
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool,
        layer: Literal["input", "hidden", "output"],
        optimizer: Literal["sgd", "adam", "adamw"],
        weight_init_std: float = 1.0,
        bias_init_std: float = 0.0,
        lr_scale: float = 1.0,
        base_width: int = 1,
    ) -> None:
        super().__init__()

        # inf dim
        if layer == "output":
            bias_width = base_width
        else:
            bias_width = out_features
        if layer == "input":
            weight_width = out_features
        else:
            weight_width = in_features

        # c = 0 for all layers by design

        # a, b, d
        if optimizer == "sgd":
            # bias scaling: Θ(1)
            bias_a = -0.5
            bias_b = 0.5
            bias_d = 0.0
            # weight
            if layer == "input":
                # scaling: Θ(1)
                weight_a = -0.5
                weight_b = 0.5
                weight_d = 0.0
            elif layer == "hidden":
                # scaling: Θ(1 / sqrt(n))
                weight_a = 0.0
                weight_b = 0.5
                weight_d = 0.0
            elif layer == "output":
                # scaling: Θ(1 / n)
                weight_a = 0.5
                weight_b = 0.5
                weight_d = 0.0

        elif optimizer in ["adam", "adamw"]:
            # bias scaling: Θ(1)
            bias_a = 0.0
            bias_b = 0.0
            bias_d = 1.0
            # weight
            if layer == "input":
                # scaling: Θ(1)
                weight_a = 0.0
                weight_b = 0.0
                weight_d = 1.0
            elif layer == "hidden":
                # scaling: Θ(1 / sqrt(n))
                weight_a = 1.0
                weight_b = -0.5
                weight_d = 2.0
            elif layer == "output":
                # scaling: Θ(1 / n)
                weight_a = 1.0
                weight_b = 0.0
                weight_d = 1.0

        else:
            raise ValueError(f"Optimizer must be either 'sgd', 'adam', or 'adamw'. Got {optimizer!r}")

        self.in_features = in_features
        self.out_features = out_features
        self.layer = layer
        self.optimizer = optimizer
        self.weight_init_std = weight_init_std
        self.bias_init_std = bias_init_std
        self.lr_scale = lr_scale
        self.base_width = base_width
        self.weight = abcdParameter(  # type: ignore[assignment]
            torch.empty(out_features, in_features),
            width=weight_width,
            a=weight_a,
            b=weight_b,
            d=weight_d,
            init_std=weight_init_std,
            lr_scale=lr_scale,
            base_width=base_width,
        )
        if bias:
            self.bias = abcdParameter(  # type: ignore[assignment]
                torch.empty(out_features),
                width=bias_width,
                a=bias_a,
                b=bias_b,
                d=bias_d,
                init_std=bias_init_std,
                lr_scale=lr_scale,
                base_width=base_width,
            )
        else:
            self.bias = abcdParameter(None)  # type: ignore[assignment]

    @staticmethod
    def scale_grad(base_width: int, width: int, d: float) -> Callable[[torch.Tensor], torch.Tensor]:
        def hook(grad: torch.Tensor) -> torch.Tensor:
            if d == 0:
                return grad
            return grad * (width / base_width) ** d

        return hook

    @property
    def weight(self) -> torch.Tensor:
        """
        The weights of the module of shape ``(out_features, in_features)``. The weight-specific learning rate and
        the initialization standard deviation are scaled with the width of the layer according to the table above.
        """
        return self.weight_multiplier * self.weight_unscaled

    @weight.setter
    def weight(self, value: abcdParameter) -> None:
        assert value.data is not None
        self.weight_multiplier = value.lr_scale**0.5 * (value.base_width / value.width) ** value.a
        self.weight_unscaled = nn.Parameter(value.data)
        if value.init_std == 0:
            self.weight_unscaled.data.zero_()
        else:
            std = value.init_std * (value.base_width / value.width) ** value.b / value.lr_scale**0.5
            self.weight_unscaled.data.normal_(mean=0.0, std=std)
        self.weight_unscaled.register_hook(self.scale_grad(value.base_width, value.width, value.d))

    @property
    def bias(self) -> torch.Tensor | None:
        """
        The bias of the module of shape ``(out_features)``. If :attr:`bias` is ``True``,
        the bias-specific learning rate and the initialization standard deviation are scaled with the width of the
        layer according to the table above.
        """
        if self.bias_unscaled is None:
            return None
        return self.bias_multiplier * self.bias_unscaled

    @bias.setter
    def bias(self, value: abcdParameter) -> None:
        self.bias_unscaled: nn.Parameter | None
        if value.data is None:
            self.register_parameter("bias_unscaled", None)
            return
        self.bias_multiplier = value.lr_scale**0.5 * (value.base_width / value.width) ** value.a
        self.bias_unscaled = nn.Parameter(value.data)
        if value.init_std == 0:
            self.bias_unscaled.data.zero_()
        else:
            std = value.init_std * (value.base_width / value.width) ** value.b / value.lr_scale**0.5
            self.bias_unscaled.data.normal_(mean=0.0, std=std)
        self.bias_unscaled.register_hook(self.scale_grad(value.base_width, value.width, value.d))

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return torch.nn.functional.linear(input, self.weight, self.bias)

    def extra_repr(self) -> str:
        return (
            f"in_features={self.in_features}, out_features={self.out_features}, bias={self.bias is not None}"
            f", layer={self.layer}, optimizer={self.optimizer}, weight_init_std={self.weight_init_std}"
            f", bias_init_std={self.bias_init_std}, lr_scale={self.lr_scale}, base_width={self.base_width}"
        )



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Sequence
from typing import Literal

import lightning.pytorch as pl
import numpy as np
import torch
import torch.distributed as dist
from lightning.pytorch.strategies import DDPStrategy

from cellarium.ml.models.model import CellariumModel
from cellarium.ml.utilities.data import get_rank_and_num_replicas
from cellarium.ml.utilities.testing import (
    assert_arrays_equal,
    assert_columns_and_array_lengths_equal,
)


class OnePassMeanVarStd(CellariumModel):
    """
    Calculate the mean, variance, and standard deviation of the data in one pass (epoch)
    using running sums and running squared sums.

    **References:**

    1. `Algorithms for calculating variance
       <https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance>`_.

    Args:
        var_names_g: The variable names schema for the input data validation.
    """

    def __init__(self, var_names_g: Sequence[str], algorithm: Literal["naive", "shifted_data"] = "naive") -> None:
        super().__init__()
        self.var_names_g = np.array(var_names_g)
        n_vars = len(self.var_names_g)
        self.n_vars = n_vars
        self.algorithm = algorithm

        self.x_sums: torch.Tensor
        self.x_squared_sums: torch.Tensor
        self.x_size: torch.Tensor
        self.x_shift: torch.Tensor | None
        self.register_buffer("x_sums", torch.empty(n_vars))
        self.register_buffer("x_squared_sums", torch.empty(n_vars))
        self.register_buffer("x_size", torch.empty(()))
        if self.algorithm == "shifted_data":
            self.register_buffer("x_shift", torch.empty(n_vars))
        else:
            self.register_buffer("x_shift", None)
        self._dummy_param = torch.nn.Parameter(torch.empty(()))
        self.reset_parameters()

    def reset_parameters(self) -> None:
        self.x_sums.zero_()
        self.x_squared_sums.zero_()
        self.x_size.zero_()
        if self.x_shift is not None:
            self.x_shift.zero_()
        self._dummy_param.data.zero_()

    def forward(self, x_ng: torch.Tensor, var_names_g: np.ndarray) -> dict[str, torch.Tensor | None]:
        """
        Args:
            x_ng:
                Gene counts matrix.
            var_names_g:
                The list of the variable names in the input data.

        Returns:
            An empty dictionary.
        """
        assert_columns_and_array_lengths_equal("x_ng", x_ng, "var_names_g", var_names_g)
        assert_arrays_equal("var_names_g", var_names_g, "var_names_g", self.var_names_g)

        if self.algorithm == "naive":
            self.x_sums = self.x_sums + x_ng.sum(dim=0)
            self.x_squared_sums = self.x_squared_sums + (x_ng**2).sum(dim=0)
            self.x_size = self.x_size + x_ng.shape[0]
        elif self.algorithm == "shifted_data":
            assert self.x_shift is not None
            if (self.x_shift == 0).all():
                _, world_size = get_rank_and_num_replicas()
                if world_size > 1:
                    gathered_x_ng = torch.zeros(
                        world_size * x_ng.shape[0], x_ng.shape[1], dtype=x_ng.dtype, device=x_ng.device
                    )
                    dist.all_gather_into_tensor(gathered_x_ng, x_ng)
                    x_shift = gathered_x_ng.mean(dim=0)
                else:
                    x_shift = x_ng.mean(dim=0)
                self.x_shift = x_shift
            self.x_sums = self.x_sums + (x_ng - self.x_shift).sum(dim=0)
            self.x_squared_sums = self.x_squared_sums + ((x_ng - self.x_shift) ** 2).sum(dim=0)
            self.x_size = self.x_size + x_ng.shape[0]
        else:
            raise ValueError(f"Unknown algorithm: {self.algorithm}")

        return {}

    def on_train_start(self, trainer: pl.Trainer) -> None:
        if trainer.world_size > 1:
            assert isinstance(
                trainer.strategy, DDPStrategy
            ), "OnePassMeanVarStd requires that the trainer uses the DDP strategy."
            assert (
                trainer.strategy._ddp_kwargs["broadcast_buffers"] is False
            ), "OnePassMeanVarStd requires that broadcast_buffers is set to False."

    def on_epoch_end(self, trainer: pl.Trainer) -> None:
        # no need to merge if only one process
        if trainer.world_size == 1:
            return

        # merge the running sums
        dist.reduce(self.x_sums, dst=0, op=dist.ReduceOp.SUM)
        dist.reduce(self.x_squared_sums, dst=0, op=dist.ReduceOp.SUM)
        dist.reduce(self.x_size, dst=0, op=dist.ReduceOp.SUM)

    @property
    def mean_g(self) -> torch.Tensor:
        """
        Mean of the data.
        """
        mean_g = self.x_sums / self.x_size
        if self.algorithm == "shifted_data":
            mean_g = mean_g + self.x_shift
        return mean_g

    @property
    def var_g(self) -> torch.Tensor:
        """
        Variance of the data.
        """
        return self.x_squared_sums / self.x_size - (self.x_sums / self.x_size) ** 2

    @property
    def std_g(self) -> torch.Tensor:
        """
        Standard deviation of the data.
        """
        return torch.sqrt(self.var_g)
